# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the subject vocabulary resource."""

from copy import deepcopy

import pytest

from invenio_vocabularies.contrib.subjects.api import Subject


@pytest.fixture(scope="module")
def prefix():
    """API prefix."""
    return "subjects"


@pytest.fixture()
def example_subject(app, db, search_clear, identity, service, subject_full_data):
    """Example subject."""
    subject = service.create(identity, subject_full_data)
    Subject.index.refresh()  # Refresh the index
    return subject


def test_get(client, h, prefix, example_subject):
    """Test the endpoint to retrieve a single item."""
    id_ = example_subject.id

    res = client.get(f"{prefix}/{id_}", headers=h)
    assert res.status_code == 200
    assert res.json["id"] == id_
    # links are encoded which seems weird
    assert res.json["links"] == {
        "self": "https://127.0.0.1:5000/api/subjects/http%3A%2F%2Fd-nb.info%2Fgnd%2F1062531973"  # noqa
    }
    # but they should still resolve
    i = res.json["links"]["self"].find("subjects")
    link = res.json["links"]["self"][i:]
    res = client.get(link, headers=h)
    assert res.status_code == 200


def test_get_invalid(client, h, prefix):
    res = client.get(f"{prefix}/invalid", headers=h)
    assert res.status_code == 404


def test_forbidden_endpoints(client, h, prefix, example_subject, subject_full_data):
    # POST
    subject_full_data_too = deepcopy(subject_full_data)
    subject_full_data_too["id"] = "other"
    res = client.post(f"{prefix}", headers=h, json=subject_full_data_too)
    assert res.status_code == 403

    # PUT
    id_ = example_subject.id
    res = client.put(f"{prefix}/{id_}", headers=h, json=subject_full_data)
    assert res.status_code == 403

    # DELETE
    res = client.delete(f"{prefix}/{id_}")
    assert res.status_code == 403


def test_search(client, h, prefix, example_subject):
    """Test a successful search."""
    res = client.get(f"{prefix}", headers=h)

    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1
    assert res.json["sortBy"] == "subject"


@pytest.fixture()
def example_subjects(app, db, search_clear, identity, service):
    subjects = [
        {
            "id": "other-1",
            "scheme": "Other",
            "title": {
                "en": "Abdomen",
            },
            "subject": "Abdomen",
        },
        {
            "id": "https://id.nlm.nih.gov/mesh/D000001",
            "scheme": "MeSH",
            "title": {
                "en": "Calcimycin",
            },
            "subject": "Calcimycin",
        },
        {
            "id": "https://id.nlm.nih.gov/mesh/D000005",
            "scheme": "MeSH",
            "title": {
                "en": "Abdomen",
            },
            "subject": "Abdomen",
        },
        {
            "id": "https://id.nlm.nih.gov/mesh/D000006",
            "scheme": "MeSH",
            "title": {
                "en": "Abdomen, Acute",
            },
            "subject": "Abdomen, Acute",
        },
        {
            "id": "yet-another-954514",
            "scheme": "Other2",
            "title": {
                "en": "Abdomen",
            },
            "subject": "Abdomen",
        },
    ]
    records = [service.create(identity, s) for s in subjects]
    Subject.index.refresh()  # Refresh the index
    return records


def test_suggest(client, h, prefix, example_subjects):
    """Test the subjects query parser."""
    # Full word
    res = client.get(f"{prefix}?suggest=abdomen", headers=h)
    assert res.json["hits"]["total"] == 4
    # The less specific text appears at the end of the list
    assert res.json["hits"]["hits"][3]["subject"] == "Abdomen, Acute"

    # Incomplete word -> finds the same results
    res = client.get(f"{prefix}?suggest=abdo", headers=h)
    assert res.json["hits"]["total"] == 4

    # Multiple words -> narrows down the results
    res = client.get(f"{prefix}?suggest=abdomen%20acute", headers=h)
    assert res.json["hits"]["total"] == 1
    assert res.json["hits"]["hits"][0]["subject"] == "Abdomen, Acute"

    # Multiple words with incomplete last word -> still finds the same result
    res = client.get(f"{prefix}?suggest=abdomen%20acu", headers=h)
    assert res.json["hits"]["total"] == 1
    assert res.json["hits"]["hits"][0]["subject"] == "Abdomen, Acute"

    # Multiple words with incomplete last word in other order -> still finds the same result
    res = client.get(f"{prefix}?suggest=acute%20abdo", headers=h)
    assert res.json["hits"]["total"] == 1
    assert res.json["hits"]["hits"][0]["subject"] == "Abdomen, Acute"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary affiliations."""

from flask_resources import JSONSerializer, ResponseHandler
from invenio_db import db
from invenio_records.dumpers import SearchDumper
from invenio_records.dumpers.indexedat import IndexedAtDumperExt
from invenio_records_resources.factories.factory import RecordTypeFactory
from invenio_records_resources.records.systemfields import ModelPIDField
from invenio_records_resources.resources.records.headers import etag_headers

from invenio_vocabularies.services.permissions import PermissionPolicy

from .config import AffiliationsSearchOptions, service_components
from .schema import AffiliationSchema

record_type = RecordTypeFactory(
    "Affiliation",
    # Data layer
    pid_field_cls=ModelPIDField,
    pid_field_kwargs={
        "model_field_name": "pid",
    },
    model_cls_attrs={
        # cannot set to nullable=False because it would fail at
        # service level when create({}), see records-resources.
        "pid": db.Column(db.String(255), unique=True),
    },
    schema_version="1.0.0",
    schema_path="local://affiliations/affiliation-v1.0.0.json",
    index_name="affiliations-affiliation-v2.0.0",
    record_dumper=SearchDumper(
        model_fields={"pid": ("id", str)},
        extensions=[
            IndexedAtDumperExt(),
        ],
    ),
    # Service layer
    service_id="affiliations",
    service_schema=AffiliationSchema,
    search_options=AffiliationsSearchOptions,
    service_components=service_components,
    permission_policy_cls=PermissionPolicy,
    # Resource layer
    endpoint_route="/affiliations",
    resource_cls_attrs={
        "response_handlers": {
            "application/json": ResponseHandler(JSONSerializer(), headers=etag_headers),
            "application/vnd.inveniordm.v1+json": ResponseHandler(
                JSONSerializer(), headers=etag_headers
            ),
        }
    },
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Localization serializer for Vocabularies."""

from functools import partial

from flask import current_app
from flask_resources import BaseListSchema, BaseObjectSchema
from invenio_i18n import get_locale
from marshmallow import fields
from marshmallow_utils.fields import BabelGettextDictField


def current_default_locale():
    """Get the Flask app's default locale."""
    if current_app:
        return current_app.config.get("BABEL_DEFAULT_LOCALE", "en")
    # Use english by default if not specified
    return "en"


L10NString = partial(BabelGettextDictField, get_locale, current_default_locale)


class VocabularyL10NItemSchema(BaseObjectSchema):
    """Vocabulary serializer schema."""

    id = fields.String(dump_only=True)
    title = L10NString(data_key="title_l10n")
    description = L10NString(data_key="description_l10n")
    props = fields.Dict(dump_only=True)
    icon = fields.String(dump_only=True)
    tags = fields.List(fields.Str(), dump_only=True)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Affiliations JSONSchema tests."""

import pytest
from jsonschema.exceptions import ValidationError

from invenio_vocabularies.contrib.affiliations.api import Affiliation


@pytest.fixture(scope="module")
def schema():
    """Returns the schema location."""
    return "local://affiliations/affiliation-v1.0.0.json"


def validates(data):
    """Validates affiliation data."""
    Affiliation(data).validate()

    return True


def fails(data):
    """Validates affiliation data."""
    pytest.raises(ValidationError, validates, data)
    return True


def test_valid_full(appctx, schema):
    data = {
        "$schema": schema,
        "acronym": "TEST",
        "identifiers": [{"identifier": "03yrm5c26", "scheme": "ror"}],
        "name": "Test affiliation",
        "title": {"en": "Test affiliation", "es": "Afiliacion de test"},
    }

    assert validates(data)


def test_valid_empty(appctx, schema):
    # check there are no requirements at JSONSchema level
    data = {"$schema": schema}

    assert validates(data)


# only acronym and name are defined by the affiliation schema
# the rest are inherited and should be tested elsewhere


def test_fails_acronym(appctx, schema):
    data = {"$schema": schema, "acronym": 1}

    assert fails(data)


def test_fails_name(appctx, schema):
    data = {"$schema": schema, "name": 1}

    assert fails(data)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Subjects services."""


class SubjectsLabels:
    """Fetching of subjects labels for facets."""

    def __call__(self, ids):
        """Return the mapping when evaluated.

        In this case, the ids received are actually the vocabulary `scheme`
        (top-level) and `subject` (nested). And since they are already
        human-readable, we keep them as-is.
        """
        unique_ids = list(set(ids))
        return {id_: id_ for id_ in unique_ids}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Transformers module."""

from abc import ABC, abstractmethod
from urllib.parse import urlparse

from lxml import etree

from .errors import TransformerError
from .xml import etree_to_dict

try:
    import rdflib
except ImportError:
    rdflib = None


class BaseTransformer(ABC):
    """Base transformer."""

    @abstractmethod
    def apply(self, stream_entry, *args, **kwargs):
        """Applies the transformation to the entry.

        :returns: A StreamEntry. The transformed entry.
                  Raises TransformerError in case of errors.
        """
        pass


class XMLTransformer(BaseTransformer):
    """XML transformer."""

    def __init__(self, root_element=None, *args, **kwargs):
        """Initializes the transformer."""
        self.root_element = root_element
        super().__init__(*args, **kwargs)

    @classmethod
    def _xml_to_etree(cls, xml):
        """Converts XML to a lxml etree."""
        return etree.HTML(xml)

    def apply(self, stream_entry, **kwargs):
        """Applies the transformation to the stream entry.

        Requires the root element to be named "record".
        """
        xml_tree = self._xml_to_etree(stream_entry.entry)
        xml_dict = etree_to_dict(xml_tree)["html"]["body"]

        if self.root_element:
            record = xml_dict.get(self.root_element)
            if not record:
                raise TransformerError(
                    f"Root element '{self.root_element}' not found in XML entry."
                )
        else:
            record = xml_dict

        stream_entry.entry = record
        return stream_entry


class RDFTransformer(BaseTransformer):
    """Base Transformer class for RDF data to dictionary format."""

    @property
    def skos_core(self):
        """Get the SKOS core namespace."""
        return rdflib.Namespace("http://www.w3.org/2004/02/skos/core#")

    def _validate_subject_url(self, subject):
        """Check if the subject is a valid URL."""
        parsed = urlparse(str(subject))
        return bool(parsed.netloc and parsed.scheme)

    def _get_identifiers(self, subject):
        """Generate identifiers field for a valid subject URL."""
        if self._validate_subject_url(subject):
            return [{"scheme": "url", "identifier": str(subject)}]
        return []

    def _get_labels(self, subject, rdf_graph):
        """Extract labels (prefLabel or altLabel) for a subject."""
        labels = {
            label.language: label.value.capitalize()
            for _, _, label in rdf_graph.triples(
                (subject, self.skos_core.prefLabel, None)
            )
            if label.language and "-" not in label.language
        }

        if "en" not in labels:
            for _, _, label in rdf_graph.triples(
                (subject, self.skos_core.altLabel, None)
            ):
                labels.setdefault(label.language, label.value.capitalize())

        return labels

    def _find_parents(self, subject, rdf_graph):
        """Find parent notations."""
        return [
            self._get_parent_notation(broader, rdf_graph)
            for broader in rdf_graph.transitive_objects(subject, self.skos_core.broader)
            if broader != subject
        ]

    def _get_parent_notation(self, broader, rdf_graph):
        """Extract notation for a parent."""
        raise NotImplementedError("This method should be implemented in a subclass.")

    def _transform_entry(self, subject, rdf_graph):
        """Transform an RDF subject entry into the desired dictionary format."""
        raise NotImplementedError("This method should be implemented in a subclass.")

    def apply(self, stream_entry, *args, **kwargs):
        """Apply transformation to a stream entry."""
        stream_entry.entry = self._transform_entry(
            stream_entry.entry["subject"], stream_entry.entry["rdf_graph"]
        )
        return stream_entry



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""EuroSciVoc subjects datastreams, readers, transformers, and writers."""

from invenio_vocabularies.datastreams.transformers import RDFTransformer

from ..config import euroscivoc_file_url


class EuroSciVocSubjectsTransformer(RDFTransformer):
    """
     Transformer class to convert EuroSciVoc RDF data to a dictionary format.

     Input:
         - Relevant fields:
             - `skos:notation`: Primary identifier for the concept.
             - `skos:prefLabel`: Preferred labels with language codes.
             - `skos:altLabel`: Alternative labels.
             - `skos:broader`: Broader concepts that this concept belongs to.

    Output:
         {
             "id": "euroscivoc:1717",  # EuroSciVoc-specific concept ID (skos:notation).
             "scheme": "EuroSciVoc",  # The scheme name indicating this is a EuroSciVoc concept.
             "subject": "Satellite radio",  # The primary subject label (first preferred label in English, skos:prefLabel).
             "title": {
                 "it": "Radio satellitare",  # Italian preferred label (skos:prefLabel).
                 "en": "Satellite radio",  # English preferred label (skos:prefLabel).
             },
             "props": {
                 "parents": "euroscivoc:1225",  # The broader concept (skos:broader), identified by its EuroSciVoc Concept ID.
             },
             "identifiers": [
                 {
                     "scheme": "url",  # Type of identifier (URL).
                     "identifier": "http://data.europa.eu/8mn/euroscivoc/87ff3577-527a-4a40-9c76-2f9d3075e2ba",  # URI of the concept (rdf:about).
                 }
             ],
         }
    """

    def _get_notation(self, subject, rdf_graph):
        """Extract the numeric notation for a subject."""
        for _, _, notation in rdf_graph.triples(
            (subject, self.skos_core.notation, None)
        ):
            if str(notation).isdigit():
                return str(notation)
        return None

    def _get_parent_notation(self, broader, rdf_graph):
        """Extract parent notation using numeric notation."""
        return self._get_notation(broader, rdf_graph)

    def _transform_entry(self, subject, rdf_graph):
        notation = self._get_notation(subject, rdf_graph)
        id = f"euroscivoc:{notation}" if notation else None
        labels = self._get_labels(subject, rdf_graph)
        parents = ",".join(
            f"euroscivoc:{n}"
            for n in reversed(self._find_parents(subject, rdf_graph))
            if n
        )

        return {
            "id": id,
            "scheme": "EuroSciVoc",
            "subject": labels.get("en", "").capitalize(),
            "title": labels,
            "props": {"parents": parents} if parents else {},
            "identifiers": self._get_identifiers(subject),
        }


# Configuration for datastream

VOCABULARIES_DATASTREAM_TRANSFORMERS = {
    "euroscivoc-transformer": EuroSciVocSubjectsTransformer
}

DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "http",
            "args": {
                "origin": euroscivoc_file_url,
            },
        },
        {
            "type": "rdf",
        },
    ],
    "transformers": [{"type": "euroscivoc-transformer"}],
    "writers": [{"args": {"writer": {"type": "subjects-service"}}, "type": "async"}],
}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Data stream factory."""

from flask import current_app

from .datastreams import DataStream
from .errors import FactoryError


class OptionsConfigMixin:
    """Options from config mixin."""

    CONFIG_VAR = None

    @classmethod
    def options(cls):
        """Reads the possible options form config."""
        return current_app.config.get(cls.CONFIG_VAR, {})


class Factory:
    """Factory."""

    FACTORY_NAME = None

    @classmethod
    def create(cls, config):
        """Creats a factory from config."""
        try:
            type_ = config["type"]
            args = config.get("args", {})
            return cls.options()[type_](**args)
        except KeyError:
            raise FactoryError(name=cls.FACTORY_NAME, key=type_)


class WriterFactory(Factory, OptionsConfigMixin):
    """Writer factory."""

    FACTORY_NAME = "Writer"
    CONFIG_VAR = "VOCABULARIES_DATASTREAM_WRITERS"


class ReaderFactory(Factory, OptionsConfigMixin):
    """Reader factory."""

    FACTORY_NAME = "Reader"
    CONFIG_VAR = "VOCABULARIES_DATASTREAM_READERS"


class TransformerFactory(Factory, OptionsConfigMixin):
    """Transformer factory."""

    FACTORY_NAME = "Transformer"
    CONFIG_VAR = "VOCABULARIES_DATASTREAM_TRANSFORMERS"


class DataStreamFactory:
    """Data streams factory."""

    @classmethod
    def create(cls, readers_config, writers_config, transformers_config=None, **kwargs):
        """Creates a data stream based on the config."""
        readers = []
        for r_conf in readers_config:
            readers.append(ReaderFactory.create(r_conf))

        writers = []
        for w_conf in writers_config:
            writers.append(WriterFactory.create(w_conf))

        transformers = []
        if transformers_config:
            for t_conf in transformers_config:
                transformers.append(TransformerFactory.create(t_conf))

        return DataStream(
            readers=readers, writers=writers, transformers=transformers, **kwargs
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Data Streams writers tests."""

from copy import deepcopy
from pathlib import Path

import pytest
import yaml

from invenio_vocabularies.datastreams import StreamEntry
from invenio_vocabularies.datastreams.errors import WriterError
from invenio_vocabularies.datastreams.writers import (
    AsyncWriter,
    ServiceWriter,
    YamlWriter,
)

##
# Service Writer
##


def test_service_writer_non_existing(lang_type, lang_data, service, identity):
    writer = ServiceWriter(service, identity=identity)
    lang = writer.write(stream_entry=StreamEntry(lang_data))
    record = service.read(identity, ("languages", lang.entry.id))
    record = record.to_dict()

    assert dict(record, **lang_data) == record


def test_service_writer_duplicate(lang_type, lang_data, service, identity):
    writer = ServiceWriter(service, identity=identity)
    _ = writer.write(stream_entry=StreamEntry(lang_data))
    with pytest.raises(WriterError) as err:
        writer.write(stream_entry=StreamEntry(lang_data))

    expected_error = [f"Vocabulary entry already exists: {lang_data}"]
    assert expected_error in err.value.args


def test_service_writer_update_existing(lang_type, lang_data, service, identity):
    # create vocabulary
    writer = ServiceWriter(service, identity=identity, update=True)
    lang = writer.write(stream_entry=StreamEntry(lang_data))
    # update vocabulary
    updated_lang = deepcopy(lang_data)
    updated_lang["description"]["en"] = "Updated english description"
    updated_lang["tags"].append("updated")
    # check changes vocabulary
    _ = writer.write(stream_entry=StreamEntry(updated_lang))
    record = service.read(identity, ("languages", lang.entry.id))
    record = record.to_dict()

    assert dict(record, **updated_lang) == record


def test_service_writer_update_non_existing(lang_type, lang_data, service, identity):
    # vocabulary item not created, call update directly
    updated_lang = deepcopy(lang_data)
    updated_lang["description"]["en"] = "Updated english description"
    updated_lang["tags"].append("updated")
    # check changes vocabulary
    writer = ServiceWriter(service, identity=identity, update=True)
    lang = writer.write(stream_entry=StreamEntry(updated_lang))
    record = service.read(identity, ("languages", lang.entry.id))
    record = record.to_dict()

    assert dict(record, **updated_lang) == record


def test_writer_wrong_config_no_insert_no_update(
    lang_type, lang_data, service, identity
):
    writer = ServiceWriter(service, identity=identity, insert=False, update=False)

    with pytest.raises(WriterError) as err:
        writer.write(stream_entry=StreamEntry(lang_data))

    expected_error = ["Writer wrongly configured to not insert and to not update"]
    assert expected_error in err.value.args


def test_writer_no_insert(lang_type, lang_data, service, identity):
    writer = ServiceWriter(service, identity=identity, insert=False, update=True)

    with pytest.raises(WriterError) as err:
        writer.write(stream_entry=StreamEntry(lang_data))

    expected_error = [f"Vocabulary entry does not exist: {lang_data}"]
    assert expected_error in err.value.args


##
# YAML Writer
##


def test_yaml_writer():
    filepath = Path("writer_test.yaml")
    test_output = [{"key_one": [{"inner_one": 1}]}, {"key_two": [{"inner_two": "two"}]}]

    writer = YamlWriter(filepath=filepath)
    for output in test_output:
        writer.write(stream_entry=StreamEntry(output))

    with open(filepath) as file:
        assert yaml.safe_load(file) == test_output

    filepath.unlink()


##
# Async Writer
##


def test_async_writer(app):
    filepath = "writer_test.yaml"
    yaml_writer_config = {"type": "yaml", "args": {"filepath": filepath}}
    async_writer = AsyncWriter(yaml_writer_config)

    test_output = [{"key_one": [{"inner_one": 1}]}, {"key_two": [{"inner_two": "two"}]}]
    for output in test_output:
        async_writer.write(stream_entry=StreamEntry(output))

    filepath = Path(filepath)
    with open(filepath) as file:
        assert yaml.safe_load(file) == test_output

    filepath.unlink()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
# Copyright (C) 2021 TU Wien.
# Copyright (C) 2024 California Institute of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

# Monkey patch Werkzeug 2.1, needed to import flask_security.login_user
# Flask-Login uses the safe_str_cmp method which has been removed in Werkzeug
# 2.1. Flask-Login v0.6.0 (yet to be released at the time of writing) fixes the
# issue. Once we depend on Flask-Login v0.6.0 as the minimal version in
# Flask-Security-Invenio/Invenio-Accounts we can remove this patch again.
try:
    # Werkzeug <2.1
    from werkzeug import security

    security.safe_str_cmp
except AttributeError:
    # Werkzeug >=2.1
    import hmac

    from werkzeug import security

    security.safe_str_cmp = hmac.compare_digest


import pytest
from flask_principal import Identity, Need, UserNeed
from flask_security import login_user
from flask_security.utils import hash_password
from invenio_access.permissions import (
    ActionUsers,
    any_user,
    superuser_access,
    system_process,
)
from invenio_access.proxies import current_access
from invenio_accounts.proxies import current_datastore
from invenio_accounts.testutils import login_user_via_session
from invenio_app.factory import create_api as _create_api
from invenio_cache import current_cache

from invenio_vocabularies.datastreams import StreamEntry
from invenio_vocabularies.records.api import Vocabulary
from invenio_vocabularies.records.models import VocabularyType

pytest_plugins = ("celery.contrib.pytest",)


@pytest.fixture(scope="module")
def h():
    """Accept JSON headers."""
    return {"accept": "application/json"}


@pytest.fixture(scope="module")
def extra_entry_points():
    """Extra entry points to load the mock_module features."""
    return {
        "invenio_db.models": [
            "mock_module = mock_module.models",
        ],
        "invenio_jsonschemas.schemas": [
            "mock_module = mock_module.jsonschemas",
        ],
        "invenio_search.mappings": [
            "records = mock_module.mappings",
        ],
    }


@pytest.fixture(scope="module")
def app_config(app_config):
    """Mimic an instance's configuration."""
    app_config["JSONSCHEMAS_HOST"] = "localhost"
    app_config["BABEL_DEFAULT_LOCALE"] = "en"
    app_config["I18N_LANGUAGES"] = [("da", "Danish")]
    app_config["RECORDS_REFRESOLVER_CLS"] = (
        "invenio_records.resolver.InvenioRefResolver"
    )
    app_config["RECORDS_REFRESOLVER_STORE"] = (
        "invenio_jsonschemas.proxies.current_refresolver_store"
    )
    return app_config


@pytest.fixture(scope="module")
def create_app(instance_path, entry_points):
    """Application factory fixture."""
    return _create_api


@pytest.fixture(scope="module")
def identity_simple():
    """Simple identity fixture."""
    i = Identity(1)
    i.provides.add(UserNeed(1))
    i.provides.add(Need(method="system_role", value="any_user"))
    return i


@pytest.fixture(scope="module")
def identity():
    """Simple identity to interact with the service."""
    i = Identity(1)
    i.provides.add(UserNeed(1))
    i.provides.add(any_user)
    i.provides.add(system_process)
    return i


@pytest.fixture(scope="module")
def superuser_identity():
    """Super user identity to interact with the services."""
    i = Identity(2)
    i.provides.add(UserNeed(2))
    i.provides.add(any_user)
    i.provides.add(system_process)
    i.provides.add(superuser_access)
    return i


@pytest.fixture(scope="module")
def service(app):
    """Vocabularies service object."""
    return app.extensions["invenio-vocabularies"].vocabularies_service


@pytest.fixture()
def lang_type(db):
    """Get a language vocabulary type."""
    v = VocabularyType.create(id="languages", pid_type="lng")
    db.session.commit()
    return v


@pytest.fixture(scope="function")
def lang_data():
    """Example data."""
    return {
        "id": "eng",
        "title": {"en": "English", "da": "Engelsk"},
        "description": {"en": "English description", "da": "Engelsk beskrivelse"},
        "icon": "file-o",
        "props": {
            "akey": "avalue",
        },
        "tags": ["recommended"],
        "type": "languages",
    }


@pytest.fixture()
def lang_data2(lang_data):
    """Example data for testing invalid cases."""
    data = dict(lang_data)
    data["id"] = "new"
    return data


@pytest.fixture()
def non_searchable_lang_data(lang_data):
    """Example data for testing unlisted cases."""
    data = dict(lang_data)
    data["tags"] = ["unlisted", "recommended"]
    return data


@pytest.fixture()
def example_record(db, identity, service, example_data):
    """Example record."""
    vocabulary_type_languages = VocabularyType(name="languages")
    vocabulary_type_licenses = VocabularyType(name="licenses")
    db.session.add(vocabulary_type_languages)
    db.session.add(vocabulary_type_licenses)
    db.session.commit()

    record = service.create(
        identity=identity,
        data=dict(**example_data, vocabulary_type_id=vocabulary_type_languages.id),
    )

    Vocabulary.index.refresh()  # Refresh the index
    return record


@pytest.fixture(scope="function")
def lang_data_many(lang_type, lang_data, service, identity):
    """Create many language vocabulary."""
    lang_ids = ["fr", "tr", "gr", "ger", "es"]
    data = dict(lang_data)

    for lang_id in lang_ids:
        data["id"] = lang_id
        service.create(identity, data)
    Vocabulary.index.refresh()  # Refresh the index
    return lang_ids


@pytest.fixture()
def user(app, db):
    """Create example user."""
    with db.session.begin_nested():
        datastore = app.extensions["security"].datastore
        _user = datastore.create_user(
            email="info@inveniosoftware.org",
            password=hash_password("password"),
            active=True,
        )
    db.session.commit()
    return _user


@pytest.fixture()
def role(app, db):
    """Create some roles."""
    with db.session.begin_nested():
        datastore = app.extensions["security"].datastore
        role = datastore.create_role(name="admin", description="admin role")

    db.session.commit()
    return role


@pytest.fixture()
def client_with_credentials(db, client, user, role):
    """Log in a user to the client."""
    current_datastore.add_role_to_user(user, role)
    action = current_access.actions["superuser-access"]
    db.session.add(ActionUsers.allow(action, user_id=user.id))

    login_user(user, remember=True)
    login_user_via_session(client, email=user.email)

    return client


@pytest.fixture()
def dict_ror_entry():
    """An example entry from ROR v2 Data Dump."""
    return StreamEntry(
        {
            "locations": [
                {
                    "geonames_id": 5381396,
                    "geonames_details": {
                        "country_code": "US",
                        "country_name": "United States",
                        "lat": 34.14778,
                        "lng": -118.14452,
                        "name": "Pasadena",
                    },
                }
            ],
            "established": 1891,
            "external_ids": [
                {
                    "type": "fundref",
                    "all": ["100006961", "100009676"],
                    "preferred": "100006961",
                },
                {
                    "type": "grid",
                    "all": ["grid.20861.3d"],
                    "preferred": "grid.20861.3d",
                },
                {"type": "isni", "all": ["0000 0001 0706 8890"], "preferred": None},
                {"type": "wikidata", "all": ["Q161562"], "preferred": None},
            ],
            "id": "https://ror.org/05dxps055",
            "domains": [],
            "links": [
                {"type": "website", "value": "http://www.caltech.edu/"},
                {
                    "type": "wikipedia",
                    "value": "http://en.wikipedia.org/wiki/California_Institute_of_Technology",
                },
            ],
            "names": [
                {"value": "CIT", "types": ["acronym"], "lang": None},
                {
                    "value": "California Institute of Technology",
                    "types": ["ror_display", "label"],
                    "lang": "en",
                },
                {"value": "Caltech", "types": ["alias"], "lang": None},
                {
                    "value": "Instituto de Tecnología de California",
                    "types": ["label"],
                    "lang": "es",
                },
            ],
            "relationships": [
                {
                    "label": "Caltech Submillimeter Observatory",
                    "type": "child",
                    "id": "https://ror.org/01e6j9276",
                },
                {
                    "label": "Infrared Processing and Analysis Center",
                    "type": "child",
                    "id": "https://ror.org/05q79g396",
                },
                {
                    "label": "Joint Center for Artificial Photosynthesis",
                    "type": "child",
                    "id": "https://ror.org/05jtgpc57",
                },
                {
                    "label": "Keck Institute for Space Studies",
                    "type": "child",
                    "id": "https://ror.org/05xkke381",
                },
                {
                    "label": "Jet Propulsion Laboratory",
                    "type": "child",
                    "id": "https://ror.org/027k65916",
                },
                {
                    "label": "Institute for Collaborative Biotechnologies",
                    "type": "child",
                    "id": "https://ror.org/04kz4p343",
                },
                {
                    "label": "Resnick Sustainability Institute",
                    "type": "child",
                    "id": "https://ror.org/04bxjes65",
                },
            ],
            "status": "active",
            "types": ["education", "funder"],
            "admin": {
                "created": {"date": "2018-11-14", "schema_version": "1.0"},
                "last_modified": {"date": "2024-05-13", "schema_version": "2.0"},
            },
        },
    )


# FIXME: https://github.com/inveniosoftware/pytest-invenio/issues/30
# Without this, success of test depends on the tests order
@pytest.fixture()
def cache():
    """Empty cache."""
    try:
        yield current_cache
    finally:
        current_cache.clear()



#
# This file is part of Invenio.
# Copyright (C) 2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create funders vocabulary table."""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.dialects import mysql, postgresql

# revision identifiers, used by Alembic.
revision = "676dd587542d"
down_revision = "17c703ce1eb7"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.create_table(
        "funder_metadata",
        sa.Column(
            "created",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        ),
        sa.Column(
            "updated",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        ),
        sa.Column("id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column(
            "json",
            sa.JSON()
            .with_variant(sqlalchemy_utils.types.json.JSONType(), "mysql")
            .with_variant(
                postgresql.JSONB(none_as_null=True, astext_type=sa.Text()), "postgresql"
            )
            .with_variant(sqlalchemy_utils.types.json.JSONType(), "sqlite"),
            nullable=True,
        ),
        sa.Column("version_id", sa.Integer(), nullable=False),
        sa.Column("pid", sa.String(255), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_funder_metadata")),
        sa.UniqueConstraint("pid", name=op.f("uq_funder_metadata_pid")),
    )
    # ### end Alembic commands ###


def downgrade():
    """Downgrade database."""
    op.drop_table("funder_metadata")
    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
# Copyright (C) 2021 Graz University of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Commands to create and manage vocabularies."""

import click
from flask.cli import with_appcontext
from invenio_access.permissions import system_identity
from invenio_pidstore.errors import PIDDeletedError, PIDDoesNotExistError

from .datastreams import DataStreamFactory
from .factories import get_vocabulary_config


@click.group()
def vocabularies():
    """Vocabularies command."""


def _process_vocab(config, num_samples=None):
    """Import a vocabulary."""
    ds = DataStreamFactory.create(
        readers_config=config["readers"],
        transformers_config=config.get("transformers"),
        writers_config=config["writers"],
        batch_size=config.get("batch_size", 1000),
        write_many=config.get("write_many", False),
    )

    success, errored, filtered = 0, 0, 0
    left = num_samples or -1
    for result in ds.process():
        left = left - 1
        if result.filtered:
            filtered += 1
        if result.errors:
            for err in result.errors:
                click.secho(err, fg="red")
            errored += 1
        else:
            success += 1
        if left == 0:
            click.secho(f"Number of samples reached {num_samples}", fg="green")
            break
    return success, errored, filtered


def _output_process(vocabulary, op, success, errored, filtered):
    """Outputs the result of an operation."""
    total = success + errored

    color = "green"
    if errored:
        color = "yellow" if success else "red"

    click.secho(
        f"Vocabulary {vocabulary} {op}. Total items {total}. \n"
        f"{success} items succeeded\n"
        f"{errored} contained errors\n"
        f"{filtered} were filtered.",
        fg=color,
    )


@vocabularies.command(name="import")
@click.option("-v", "--vocabulary", type=click.STRING, required=True)
@click.option("-f", "--filepath", type=click.STRING)
@click.option("-o", "--origin", type=click.STRING)
@click.option("-n", "--num-samples", type=click.INT)
@with_appcontext
def import_vocab(vocabulary, filepath=None, origin=None, num_samples=None):
    """Import a vocabulary."""
    if not filepath and not origin:
        click.secho("One of --filepath or --origin must be present.", fg="red")
        exit(1)

    vc = get_vocabulary_config(vocabulary)
    config = vc.get_config(filepath, origin)

    success, errored, filtered = _process_vocab(config, num_samples)

    _output_process(vocabulary, "imported", success, errored, filtered)


@vocabularies.command()
@click.option("-v", "--vocabulary", type=click.STRING, required=True)
@click.option("-f", "--filepath", type=click.STRING)
@click.option("-o", "--origin", type=click.STRING)
@with_appcontext
def update(vocabulary, filepath=None, origin=None):
    """Import a vocabulary."""
    if not filepath and not origin:
        click.secho("One of --filepath or --origin must be present.", fg="red")
        exit(1)
    vc = get_vocabulary_config(vocabulary)
    config = vc.get_config(filepath, origin)

    for w_conf in config["writers"]:
        if w_conf["type"] == "async":
            w_conf_update = w_conf["args"]["writer"]
        else:
            w_conf_update = w_conf

        if "args" in w_conf_update:
            w_conf_update["args"]["update"] = True
        else:
            w_conf_update["args"] = {"update": True}

    success, errored, filtered = _process_vocab(config)

    _output_process(vocabulary, "updated", success, errored, filtered)


@vocabularies.command()
@click.option("-v", "--vocabulary", type=click.STRING, required=True)
@click.option("-f", "--filepath", type=click.STRING)
@click.option("-o", "--origin", type=click.STRING)
@click.option("-t", "--target", type=click.STRING)
@click.option("-n", "--num-samples", type=click.INT)
@with_appcontext
def convert(vocabulary, filepath=None, origin=None, target=None, num_samples=None):
    """Convert a vocabulary to a new format."""
    if not filepath and (not origin or not target):
        click.secho(
            "One of --filepath or --origin and --target must be present.", fg="red"
        )
        exit(1)

    vc = get_vocabulary_config(vocabulary)
    config = vc.get_config(filepath, origin)
    if not filepath:
        config["writers"] = [{"type": "yaml", "args": {"filepath": target}}]

    success, errored, filtered = _process_vocab(config, num_samples)
    _output_process(vocabulary, "converted", success, errored, filtered)


@vocabularies.command()
@click.option("-v", "--vocabulary", type=click.STRING, required=True)
@click.option(
    "-i",
    "--identifier",
    type=click.STRING,
    help="Identifier of the vocabulary item to delete.",
)
@click.option("--all", is_flag=True, default=False)
@with_appcontext
def delete(vocabulary, identifier, all):
    """Delete all items or a specific one of the vocabulary."""
    if not identifier and not all:
        click.secho("An identifier or the --all flag must be present.", fg="red")
        exit(1)

    vc = get_vocabulary_config(vocabulary)
    service = vc.get_service()
    if identifier:
        try:
            if service.delete(system_identity, identifier):
                click.secho(f"{identifier} deleted from {vocabulary}.", fg="green")
        except (PIDDeletedError, PIDDoesNotExistError):
            click.secho(f"PID {identifier} not found.")
    elif all:
        items = service.scan(system_identity)
        for item in items.hits:
            try:
                if service.delete(system_identity, item["id"]):
                    click.secho(f"{item['id']} deleted from {vocabulary}.", fg="green")
            except (PIDDeletedError, PIDDoesNotExistError):
                click.secho(f"PID {item['id']} not found.")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Subject JSONSchema tests."""

import pytest
from jsonschema.exceptions import ValidationError

from invenio_vocabularies.contrib.subjects.api import Subject


@pytest.fixture(scope="module")
def schema():
    """Returns the schema location."""
    return "local://subjects/subject-v1.0.0.json"


def validates(data):
    """Validates subject data."""
    Subject(data).validate()
    return True


def fails(data):
    """Validates affiliation data."""
    pytest.raises(ValidationError, validates, data)
    return True


def test_valid_full(appctx, schema):
    data = {
        "$schema": schema,
        "title": {
            "en": "Dark Web",
            "de": "Darknet",
            "fr": "Réseaux anonymes (informatique)",
        },
        "id": "http://d-nb.info/gnd/1062531973",
        "pid": {"pk": 1, "status": "R", "pid_type": "subid", "obj_type": "sub"},
        "scheme": "GND",
        "synonyms": ["Deep Web"],
    }
    assert validates(data)


def test_valid_empty(appctx, schema):
    # check there are no requirements at JSONSchema level
    data = {"$schema": schema}

    assert validates(data)


def test_fails_scheme(appctx, schema):
    data = {"$schema": schema, "scheme": 1}

    assert fails(data)


def test_fails_name(appctx, schema):
    data = {"$schema": schema, "subject": 1}

    assert fails(data)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary resource schema."""

from marshmallow import Schema, fields

from invenio_vocabularies.resources.serializer import L10NString


class VocabularyL10Schema(Schema):
    """Vocabulary schema."""

    id = fields.String()
    title = L10NString(data_key="title_l10n")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2024 University of Münster.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

import pytest
from invenio_records_resources.proxies import current_service_registry


@pytest.fixture(scope="function")
def subject_full_data():
    """Controlled vocabulary backed subject data."""
    return {
        "title": {
            "en": "Dark Web",
            "de": "Darknet",
            "fr": "Réseaux anonymes (informatique)",
        },
        "subject": "Darknet",
        "id": "http://d-nb.info/gnd/1062531973",
        "scheme": "GND",
        "synonyms": ["Deep Web"],
    }


@pytest.fixture(scope="function")
def expected_subject_full_data():
    """Controlled vocabulary backed subject data."""
    return {
        "title": {
            "en": "Dark Web",
            "de": "Darknet",
            "fr": "Réseaux anonymes (informatique)",
        },
        "subject": "Darknet",
        "id": "http://d-nb.info/gnd/1062531973",
        "scheme": "GND",
        "synonyms": ["Deep Web"],
    }


@pytest.fixture(scope="module")
def service():
    """Subjects service object."""
    return current_service_registry.get("subjects")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Affiliations module."""

from .resources import AffiliationsResource, AffiliationsResourceConfig
from .services import AffiliationsService, AffiliationsServiceConfig

__all__ = [
    "AffiliationsResource",
    "AffiliationsResourceConfig",
    "AffiliationsService",
    "AffiliationsServiceConfig",
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Jobs module."""

import datetime

from invenio_i18n import gettext as _
from invenio_jobs.jobs import JobType

from invenio_vocabularies.services.tasks import process_datastream

from .contrib.names.datastreams import ORCID_PRESET_DATASTREAM_CONFIG


class ProcessDataStreamJob(JobType):
    """Generic process data stream job type."""

    task = process_datastream


class ProcessRORAffiliationsJob(ProcessDataStreamJob):
    """Process ROR affiliations datastream registered task."""

    description = "Process ROR affiliations"
    title = "Load ROR affiliations"
    id = "process_ror_affiliations"

    @classmethod
    def build_task_arguments(cls, job_obj, since=None, **kwargs):
        """Process ROR affiliations."""
        # NOTE: Update is set to False for now given we don't have the logic to re-index dependent records yet.
        # Since jobs support custom args, update true can be passed via that.
        return {
            "config": {
                "readers": [
                    {
                        "args": {"since": str(since)},
                        "type": "ror-http",
                    },
                    {"args": {"regex": "_schema_v2\\.json$"}, "type": "zip"},
                    {"type": "json"},
                ],
                "writers": [
                    {
                        "args": {
                            "writer": {
                                "type": "affiliations-service",
                                "args": {"update": False},
                            }
                        },
                        "type": "async",
                    }
                ],
                "transformers": [{"type": "ror-affiliations"}],
            }
        }


class ProcessRORFundersJob(ProcessDataStreamJob):
    """Process ROR funders datastream registered task."""

    description = "Process ROR funders"
    title = "Load ROR funders"
    id = "process_ror_funders"

    @classmethod
    def build_task_arguments(cls, job_obj, since=None, **kwargs):
        """Process ROR funders."""
        # NOTE: Update is set to False for now given we don't have the logic to re-index dependent records yet.
        # Since jobs support custom args, update true can be passed via that.
        return {
            "config": {
                "readers": [
                    {
                        "args": {"since": str(since)},
                        "type": "ror-http",
                    },
                    {"args": {"regex": "_schema_v2\\.json$"}, "type": "zip"},
                    {"type": "json"},
                ],
                "writers": [
                    {
                        "args": {
                            "writer": {
                                "type": "funders-service",
                                "args": {"update": False},
                            }
                        },
                        "type": "async",
                    }
                ],
                "transformers": [{"type": "ror-funders"}],
            }
        }


class ImportAwardsOpenAIREJob(ProcessDataStreamJob):
    """Import awards from OpenAIRE registered task."""

    description = "Import awards from OpenAIRE"
    title = "Import Awards OpenAIRE"
    id = "import_awards_openaire"

    @classmethod
    def build_task_arguments(cls, job_obj, since=None, **kwargs):
        """Process awards OpenAIRE."""
        return {
            "config": {
                "readers": [
                    {
                        "type": "openaire-http",
                        "args": {"origin": "diff", "tar_href": "/project.tar"},
                    },
                    {
                        "type": "tar",
                        "args": {
                            "mode": "r",
                            "regex": "\\.json.gz$",
                        },
                    },
                    {"type": "gzip"},
                    {"type": "jsonl"},
                ],
                "transformers": [{"type": "openaire-award"}],
                "writers": [
                    {"args": {"writer": {"type": "awards-service"}}, "type": "async"}
                ],
            }
        }


class UpdateAwardsCordisJob(ProcessDataStreamJob):
    """Update awards from CORDIS registered task."""

    description = "Update awards from CORDIS"
    title = "Update Awards CORDIS"
    id = "update_awards_cordis"

    @classmethod
    def build_task_arguments(cls, job_obj, since=None, **kwargs):
        """Process awards Cordis."""
        return {
            "config": {
                "readers": [
                    {"args": {"origin": "HE"}, "type": "cordis-project-http"},
                    {"args": {"mode": "r", "regex": "\\.xml$"}, "type": "zip"},
                    {"args": {"root_element": "project"}, "type": "xml"},
                ],
                "transformers": [{"type": "cordis-award"}],
                "writers": [
                    {
                        "args": {"writer": {"type": "cordis-awards-service"}},
                        "type": "async",
                    }
                ],
            }
        }


class ImportORCIDJob(ProcessDataStreamJob):
    """Import ORCID data registered task."""

    description = "Import ORCID data"
    title = "Import ORCID data"
    id = "import_orcid"

    @classmethod
    def build_task_arguments(cls, job_obj, since=None, **kwargs):
        """Process ORCID data."""
        return {"config": {**ORCID_PRESET_DATASTREAM_CONFIG}}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""OpenAIRE-related Datastreams Readers/Writers/Transformers tests."""

import io
from unittest.mock import patch

import pytest

from invenio_vocabularies.contrib.common.openaire.datastreams import OpenAIREHTTPReader
from invenio_vocabularies.datastreams.errors import ReaderError

API_JSON_RESPONSE_CONTENT = {
    "linkset": [
        {
            "anchor": "https://example.com/records/10488385",
            "item": [
                {
                    "href": "https://example.com/records/10488385/files/organization.tar",
                    "type": "application/x-tar",
                },
                {
                    "href": "https://example.com/records/10488385/files/project.tar",
                    "type": "application/x-tar",
                },
            ],
        },
        {
            "anchor": "https://example.com/api/records/10488385",
            "describes": [
                {"href": "https://example.com/records/10488385", "type": "text/html"}
            ],
            "type": "application/dcat+xml",
        },
    ]
}

API_JSON_RESPONSE_CONTENT_WRONG_NUMBER_PROJECT_TAR_ITEMS_ERROR = {
    "linkset": [
        {
            "anchor": "https://example.com/records/10488385",
            "item": [
                {
                    "href": "https://example.com/records/10488385/files/organization.tar",
                    "type": "application/x-tar",
                },
                {
                    "href": "https://example.com/records/10488385/files/project.tar",
                    "type": "application/x-tar",
                },
                {
                    "href": "https://example.com/another/project.tar",
                    "type": "application/x-tar",
                },
            ],
        },
        {
            "anchor": "https://example.com/api/records/10488385",
            "describes": [
                {"href": "https://example.com/records/10488385", "type": "text/html"}
            ],
            "type": "application/dcat+xml",
        },
    ]
}

DOWNLOAD_FILE_BYTES_CONTENT = b"The content of the file"


class MockResponse:
    content = DOWNLOAD_FILE_BYTES_CONTENT

    def __init__(self, api_json_response_content):
        self.api_json_response_content = api_json_response_content

    def json(self, **kwargs):
        return self.api_json_response_content

    def raise_for_status(self):
        pass


@pytest.fixture(scope="function")
def download_file_bytes_content():
    return DOWNLOAD_FILE_BYTES_CONTENT


@patch(
    "requests.get",
    side_effect=lambda url, headers=None: MockResponse(API_JSON_RESPONSE_CONTENT),
)
def test_openaire_http_reader(_, download_file_bytes_content):
    reader = OpenAIREHTTPReader(origin="full", tar_href="/project.tar")
    results = []
    for entry in reader.read():
        results.append(entry)

    assert len(results) == 1
    assert isinstance(results[0], io.BytesIO)
    assert results[0].read() == download_file_bytes_content


@patch(
    "requests.get",
    side_effect=lambda url, headers=None: MockResponse(
        API_JSON_RESPONSE_CONTENT_WRONG_NUMBER_PROJECT_TAR_ITEMS_ERROR
    ),
)
def test_openaire_http_reader_wrong_number_tar_items_error(_):
    reader = OpenAIREHTTPReader(origin="full", tar_href="/project.tar")
    with pytest.raises(ReaderError):
        next(reader.read())


def test_openaire_http_reader_unsupported_origin_option():
    reader = OpenAIREHTTPReader(origin="unsupported_origin_option")
    with pytest.raises(ReaderError):
        next(reader.read())


def test_openaire_http_reader_item_not_implemented():
    reader = OpenAIREHTTPReader()
    with pytest.raises(NotImplementedError):
        next(reader.read("A fake item"))


def test_openaire_http_reader_iter_not_implemented():
    reader = OpenAIREHTTPReader()
    with pytest.raises(NotImplementedError):
        reader._iter("A fake file pointer")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Data Streams tasks tests."""

from pathlib import Path

import yaml

from invenio_vocabularies.datastreams.tasks import write_entry, write_many_entry


def test_write_entry(app):
    filepath = "writer_test.yaml"
    yaml_writer_config = {"type": "yaml", "args": {"filepath": filepath}}
    entry = {"key_one": [{"inner_one": 1}]}
    write_entry(yaml_writer_config, entry)

    filepath = Path(filepath)
    with open(filepath) as file:
        assert yaml.safe_load(file) == [entry]

    filepath.unlink()


def test_write_many_entry(app):
    filepath = "writer_test.yaml"
    yaml_writer_config = {"type": "yaml", "args": {"filepath": filepath}}
    entries = [{"key_one": [{"inner_one": 1}]}, {"key_two": [{"inner_two": 2}]}]
    write_many_entry(yaml_writer_config, entries)

    filepath = Path(filepath)
    with open(filepath) as file:
        assert yaml.safe_load(file) == entries

    filepath.unlink()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names vocabulary resources."""

from flask import g
from flask_resources import resource_requestctx, response_handler, route
from invenio_records_resources.resources.records.resource import request_view_args
from marshmallow import fields

from .names import record_type


class NamesResourceConfig(record_type.resource_config_cls):
    """Name resource."""

    routes = record_type.resource_config_cls.routes
    routes["item-names-resolve"] = "/<pid_type>/<pid_value>"
    request_view_args = {
        "pid_value": fields.Str(),
        "pid_type": fields.Str(),
    }


class NamesResource(record_type.resource_cls):
    """Name resource."""

    def create_url_rules(self):
        """Create the URL rules for the record resource."""
        routes = self.config.routes
        url_rules = super(NamesResource, self).create_url_rules()
        url_rules += [
            route("GET", routes["item-names-resolve"], self.name_resolve_by_id),
        ]

        return url_rules

    @request_view_args
    @response_handler()
    def name_resolve_by_id(self):
        """Resolve an identifier."""
        item = self.service.resolve(
            id_=resource_requestctx.view_args["pid_value"],
            id_type=resource_requestctx.view_args["pid_type"],
            identity=g.identity,
        )

        return item.to_dict(), 200



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Subjects services."""

from invenio_db import db

from ...records.models import VocabularyScheme
from .subjects import record_type

SubjectsServiceConfig = record_type.service_config_cls


class SubjectsService(record_type.service_cls):
    """Subjects service."""

    def create_scheme(self, identity, id_, name="", uri=""):
        """Create a row for the subject scheme metadata."""
        self.require_permission(identity, "manage")
        scheme = VocabularyScheme.create(
            id=id_, parent_id="subjects", name=name, uri=uri
        )
        db.session.commit()
        return scheme



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Awards datastreams, transformers, writers and readers."""

import io

import requests
from flask import current_app
from invenio_access.permissions import system_identity
from invenio_i18n import lazy_gettext as _

from invenio_vocabularies.datastreams.errors import ReaderError

from ...datastreams.errors import TransformerError
from ...datastreams.readers import BaseReader
from ...datastreams.transformers import BaseTransformer
from ...datastreams.writers import ServiceWriter
from .config import awards_ec_ror_id, awards_openaire_funders_mapping


class AwardsServiceWriter(ServiceWriter):
    """Funders service writer."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        service_or_name = kwargs.pop("service_or_name", "awards")
        super().__init__(service_or_name=service_or_name, *args, **kwargs)

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return entry["id"]


class OpenAIREProjectTransformer(BaseTransformer):
    """Transforms an OpenAIRE project record into an award record."""

    def apply(self, stream_entry, **kwargs):
        """Applies the transformation to the stream entry."""
        record = stream_entry.entry
        award = {}

        code = record["code"]

        # The `id` should follow the format `sourcePrefix::md5(localId)` where `sourcePrefix` is 12 characters long.
        # See: https://graph.openaire.eu/docs/data-model/pids-and-identifiers#identifiers-in-the-graph
        #
        # The format of `id` in the full OpenAIRE Graph Dataset (https://doi.org/10.5281/zenodo.3516917)
        # follows this format (e.g. 'abc_________::0123456789abcdef0123456789abcdef').
        # However, the format of `id` in the new collected projects dataset (https://doi.org/10.5281/zenodo.6419021)
        # does not follow this format, and has a `40|` prefix (e.g. '40|abc_________::0123456789abcdef0123456789abcdef').
        #
        # The number '40' corresponds to the entity types 'Project'.
        # See: https://ec.europa.eu/research/participants/documents/downloadPublic?documentIds=080166e5a3a1a213&appId=PPGMS
        # See: https://graph.openaire.eu/docs/5.0.0/data-model/entities/project#id
        openaire_funder_prefix = record["id"].split("::", 1)[0].split("|", 1)[-1]

        funder_id = awards_openaire_funders_mapping.get(openaire_funder_prefix)
        if funder_id is None:
            raise TransformerError(
                _(
                    "Unknown OpenAIRE funder prefix {openaire_funder_prefix}".format(
                        openaire_funder_prefix=openaire_funder_prefix
                    )
                )
            )

        award["id"] = f"{funder_id}::{code}"

        funding = next(iter(record.get("funding", [])), None)
        if funding:
            funding_stream_id = funding.get("fundingStream", {}).get("id", "")
            # Example funding stream ID: `EC::HE::HORIZON-AG-UN`. We need the `HE`
            # string, i.e. the second "part" of the identifier.
            program = next(iter(funding_stream_id.split("::")[1:2]), "")
            if program:
                award["program"] = program

        identifiers = []
        if funder_id == awards_ec_ror_id:
            identifiers.append(
                {
                    "identifier": f"https://cordis.europa.eu/projects/{code}",
                    "scheme": "url",
                }
            )
        elif record.get("websiteurl"):
            identifiers.append(
                {"identifier": record.get("websiteurl"), "scheme": "url"}
            )

        if identifiers:
            award["identifiers"] = identifiers

        award["number"] = code

        # `title` is a mandatory attribute of the `Project` object in the OpenAIRE Graph Data Model.
        # See: https://graph.openaire.eu/docs/data-model/entities/project#title
        # However, 15'000+ awards for the FCT funder (and 1 record the NIH funder) are missing a title attribute.
        if "title" not in record:
            raise TransformerError(
                _(
                    "Missing title attribute for award {award_id}".format(
                        award_id=award["id"]
                    )
                )
            )
        award["title"] = {"en": record["title"]}

        award["funder"] = {"id": funder_id}
        acronym = record.get("acronym")
        if acronym:
            award["acronym"] = acronym

        if "startDate" in record:
            award["start_date"] = record["startDate"]
        if "endDate" in record:
            award["end_date"] = record["endDate"]
        if "summary" in record:
            award["description"] = {"en": record["summary"]}

        stream_entry.entry = award
        return stream_entry


class CORDISProjectHTTPReader(BaseReader):
    """CORDIS Project HTTP Reader returning an in-memory binary stream of the latest CORDIS Horizon Europe project zip file."""

    def _iter(self, fp, *args, **kwargs):
        raise NotImplementedError(
            "CORDISProjectHTTPReader downloads one file and therefore does not iterate through items"
        )

    def read(self, item=None, *args, **kwargs):
        """Reads the latest CORDIS Horizon Europe project zip file and yields an in-memory binary stream of it."""
        if item:
            raise NotImplementedError(
                "CORDISProjectHTTPReader does not support being chained after another reader"
            )

        if self._origin == "HE":
            file_url = "https://cordis.europa.eu/data/cordis-HORIZONprojects-xml.zip"
        elif self._origin == "H2020":
            file_url = "https://cordis.europa.eu/data/cordis-h2020projects-xml.zip"
        elif self._origin == "FP7":
            file_url = "https://cordis.europa.eu/data/cordis-fp7projects-xml.zip"
        else:
            raise ReaderError(
                "The --origin option should be either 'HE' (for Horizon Europe) or 'H2020' (for Horizon 2020) or 'FP7'"
            )

        # Download the ZIP file and fully load the response bytes content in memory.
        # The bytes content are then wrapped by a BytesIO to be file-like object (as required by `zipfile.ZipFile`).
        # Using directly `file_resp.raw` is not possible since `zipfile.ZipFile` requires the file-like object to be seekable.
        file_resp = requests.get(file_url)
        file_resp.raise_for_status()
        yield io.BytesIO(file_resp.content)


class CORDISProjectTransformer(BaseTransformer):
    """Transforms a CORDIS project record into an award record."""

    def apply(self, stream_entry, **kwargs):
        """Applies the transformation to the stream entry."""
        record = stream_entry.entry
        award = {}

        # Here `id` is the project ID, which will be used to attach the update to the existing project.
        award["id"] = (
            f"{current_app.config['VOCABULARIES_AWARDS_EC_ROR_ID']}::{record['id']}"
        )

        categories = record.get("relations", {}).get("categories", {}).get("category")
        if categories:
            if isinstance(categories, dict):
                categories = [categories]

            award["subjects"] = [
                {"id": f"euroscivoc:{vocab_id}"}
                for category in categories
                if category.get("@classification") == "euroSciVoc"
                and (vocab_id := category["code"].split("/")[-1]).isdigit()
            ]

        organizations = (
            record.get("relations", {}).get("associations", {}).get("organization")
        )
        if organizations:
            # Projects with a single organization are not wrapped in a list,
            # so we do this here to be able to iterate over it.
            organizations = (
                organizations if isinstance(organizations, list) else [organizations]
            )
            award["organizations"] = []
            for organization in organizations:
                # Some organizations in FP7 projects do not have a "legalname" key,
                # for instance the 14th participant in "SAGE" https://cordis.europa.eu/project/id/999902.
                # In this case, fully skip the organization entry.
                if "legalname" not in organization:
                    continue

                organization_data = {
                    "organization": organization["legalname"],
                }

                # Some organizations in FP7 projects do not have an "id" key (the PIC identifier),
                # for instance "AIlGreenVehicles" in "MOTORBRAIN" https://cordis.europa.eu/project/id/270693.
                # In this case, still store the name but skip the identifier part.
                if "id" in organization:
                    organization_data.update(
                        {
                            "scheme": "pic",
                            "id": organization["id"],
                        }
                    )

                award["organizations"].append(organization_data)

        programmes = (
            record.get("relations", {}).get("associations", {}).get("programme", {})
        )
        if programmes:
            # Projects with a single programme (this is the case of some projects in FP7) are not wrapped in a list,
            # so we do this here to be able to iterate over it.
            programmes = programmes if isinstance(programmes, list) else [programmes]

            programmes_related_legal_basis = [
                {
                    "code": programme["code"],
                    "uniqueprogrammepart": programme.get("@uniqueprogrammepart"),
                }
                for programme in programmes
                if programme.get("@type") == "relatedLegalBasis"
            ]

            if len(programmes_related_legal_basis) == 0:
                raise TransformerError(
                    _(
                        "No related legal basis programme found for project {project_id}".format(
                            project_id=record["id"]
                        )
                    )
                )
            elif len(programmes_related_legal_basis) == 1:
                # FP7 projects have only one related legal basis programme and do not have a 'uniqueprogrammepart' field.
                unique_programme_related_legal_basis = programmes_related_legal_basis[0]
            elif len(programmes_related_legal_basis) >= 1:
                # The entry with the field 'uniqueprogrammepart' == 'true' is the high level programme code,
                # while the other entry is a more specific sub-programme.
                unique_programme_related_legal_basis = [
                    programme_related_legal_basis
                    # A few H2020 projects have more than one 'uniqueprogrammepart' == 'true',
                    # for instance https://cordis.europa.eu/project/id/825673 (showing as "main programme" in the page)
                    # which has one entry with the code 'H2020-EU.1.2.',
                    # and one with the code 'H2020-EU.1.2.3.'.
                    # We sort them from the shortest code to the longest code, and take the first item,
                    # so that it conforms more with other projects which all have the shortest code as the main one.
                    for programme_related_legal_basis in sorted(
                        programmes_related_legal_basis, key=lambda d: len(d["code"])
                    )
                    if programme_related_legal_basis["uniqueprogrammepart"] == "true"
                ][0]

            # Store the code of the programme.
            # For instance the code "HORIZON.1.2" which means "Marie Skłodowska-Curie Actions (MSCA)"
            # See https://cordis.europa.eu/programme/id/HORIZON.1.2
            award["program"] = unique_programme_related_legal_basis["code"]

        stream_entry.entry = award
        return stream_entry


class CORDISAwardsServiceWriter(ServiceWriter):
    """CORDIS Awards service writer."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        service_or_name = kwargs.pop("service_or_name", "awards")
        # Here we only update and we do not insert, since CORDIS data is used to augment existing awards
        # (with subjects, organizations, and program information) and is not used to create new awards.
        super().__init__(
            service_or_name=service_or_name, insert=False, update=True, *args, **kwargs
        )

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return entry["id"]


VOCABULARIES_DATASTREAM_READERS = {
    "cordis-project-http": CORDISProjectHTTPReader,
}

VOCABULARIES_DATASTREAM_TRANSFORMERS = {
    "openaire-award": OpenAIREProjectTransformer,
    "cordis-award": CORDISProjectTransformer,
}
"""ORCiD Data Streams transformers."""

VOCABULARIES_DATASTREAM_WRITERS = {
    "awards-service": AwardsServiceWriter,
    "cordis-awards-service": CORDISAwardsServiceWriter,
}
"""ORCiD Data Streams transformers."""

DATASTREAM_CONFIG_CORDIS = {
    "readers": [
        {"type": "cordis-project-http"},
        {
            "type": "zip",
            "args": {
                "regex": "\\.xml$",
                "mode": "r",
            },
        },
        {
            "type": "xml",
            "args": {
                "root_element": "project",
            },
        },
    ],
    "transformers": [
        {"type": "cordis-award"},
    ],
    "writers": [
        {
            "type": "cordis-awards-service",
            "args": {
                "identity": system_identity,
            },
        }
    ],
}
"""Data Stream configuration.

An origin is required for the reader.
"""

DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "tar",
            "args": {
                "regex": "\\.json.gz$",
                "mode": "r",
            },
        },
        {"type": "gzip"},
        {"type": "jsonl"},
    ],
    "transformers": [
        {"type": "openaire-award"},
    ],
    "writers": [
        {
            "type": "awards-service",
            "args": {
                "identity": system_identity,
            },
        }
    ],
}
"""Data Stream configuration.

An origin is required for the reader.
"""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Base data stream."""

from flask import current_app

from .errors import ReaderError, TransformerError, WriterError


class StreamEntry:
    """Object to encapsulate streams processing."""

    def __init__(self, entry, record=None, errors=None, op_type=None, exc=None):
        """Constructor for the StreamEntry class.

        :param entry (object): The entry object, usually a record dict.
        :param record (object): The record object, usually a record class.
        :param errors (list, optional): List of errors. Defaults to None.
        :param op_type (str, optional): The operation type. Defaults to None.
        :param exc (str, optional): The raised unhandled exception. Defaults to None.
        """
        self.entry = entry
        self.record = record
        self.filtered = False
        self.errors = errors or []
        self.op_type = op_type
        self.exc = exc

    def log_errors(self, logger=None):
        """Log the errors using the provided logger or the default logger.

        :param logger (logging.Logger, optional): Logger instance to use. Defaults to None.
        """
        if logger is None:
            logger = current_app.logger
        for error in self.errors:
            logger.error(f"Error in entry {self.entry}: {error}")
        if self.exc:
            logger.error(f"Exception in entry {self.entry}: {self.exc}")


class DataStream:
    """Data stream."""

    def __init__(
        self,
        readers,
        writers,
        transformers=None,
        batch_size=100,
        write_many=False,
        *args,
        **kwargs,
    ):
        """Constructor.

        :param readers: an ordered list of readers.
        :param writers: an ordered list of writers.
        :param transformers: an ordered list of transformers to apply.
        """
        self._readers = readers
        self._transformers = transformers
        self._writers = writers
        self.batch_size = batch_size
        self.write_many = write_many

    def filter(self, stream_entry, *args, **kwargs):
        """Checks if an stream_entry should be filtered out (skipped)."""
        return False

    def process_batch(self, batch):
        """Process a batch of entries."""
        transformed_entries = []
        for stream_entry in batch:
            if stream_entry.errors:
                yield stream_entry  # reading errors
            else:
                transformed_entry = self.transform(stream_entry)
                if transformed_entry.errors:
                    yield transformed_entry
                elif self.filter(transformed_entry):
                    transformed_entry.filtered = True
                    yield transformed_entry
                else:
                    transformed_entries.append(transformed_entry)
        if transformed_entries:
            if self.write_many:
                yield from self.batch_write(transformed_entries)
            else:
                yield from (self.write(entry) for entry in transformed_entries)

    def process(self, *args, **kwargs):
        """Iterates over the entries.

        Uses the reader to get the raw entries and transforms them.
        It will iterate over the `StreamEntry` objects returned by
        the reader, apply the transformations and yield the result of
        writing it.
        """
        batch = []
        for stream_entry in self.read():
            batch.append(stream_entry)
            if len(batch) >= self.batch_size:
                yield from self.process_batch(batch)
                batch = []

        # Process any remaining entries in the last batch
        if batch:
            yield from self.process_batch(batch)

    def read(self):
        """Recursively read the entries."""

        def pipe_gen(gen_funcs, piped_item=None):
            _gen_funcs = list(gen_funcs)  # copy to avoid modifying ref list
            # use and remove the current generator
            current_gen_func = _gen_funcs.pop(0)
            for item in current_gen_func(piped_item):
                try:
                    # exhaust iterations of subsequent generators
                    if _gen_funcs:
                        yield from pipe_gen(_gen_funcs, piped_item=item)
                    # there is no subsequent generator, return the current item
                    else:
                        yield StreamEntry(item)
                except ReaderError as err:
                    yield StreamEntry(
                        entry=item,
                        errors=[f"{current_gen_func.__qualname__}: {str(err)}"],
                    )

        read_gens = [r.read for r in self._readers]
        yield from pipe_gen(read_gens)

    def transform(self, stream_entry, *args, **kwargs):
        """Apply the transformations to an stream_entry."""
        for transformer in self._transformers:
            try:
                stream_entry = transformer.apply(stream_entry)
            except TransformerError as err:
                stream_entry.errors.append(
                    f"{transformer.__class__.__name__}: {str(err)}"
                )
                return stream_entry  # break loop

        return stream_entry

    def write(self, stream_entry, *args, **kwargs):
        """Apply the transformations to an stream_entry."""
        for writer in self._writers:
            try:
                writer.write(stream_entry)
            except WriterError as err:
                stream_entry.errors.append(f"{writer.__class__.__name__}: {str(err)}")

        return stream_entry

    def batch_write(self, stream_entries, *args, **kwargs):
        """Apply the transformations to an stream_entry. Errors are handler in the service layer."""
        for writer in self._writers:
            yield from writer.write_many(stream_entries)

    def total(self, *args, **kwargs):
        """The total of entries obtained from the origin."""
        raise NotImplementedError()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary components."""

from invenio_i18n import lazy_gettext as _
from invenio_records_resources.services.records.components import ServiceComponent
from marshmallow import ValidationError
from sqlalchemy.orm.exc import NoResultFound

from ..records.models import VocabularyType


class VocabularyTypeComponent(ServiceComponent):
    """Set the record's vocabulary type."""

    def _set_type(self, data, record):
        type_id = data.pop("type", None)
        if type_id:
            try:
                record.type = VocabularyType.query.filter_by(id=type_id["id"]).one()
            except NoResultFound:
                raise ValidationError(
                    _("The vocabulary type does not exists."),
                    field_name="type",
                )

    def create(self, identity, data=None, record=None, **kwargs):
        """Inject vocabulary type to the record."""
        self._set_type(data, record)

    def update(self, identity, data=None, record=None, **kwargs):
        """Inject vocabulary type to the record."""
        self._set_type(data, record)


class PIDComponent(ServiceComponent):
    """PID registration component."""

    def create(self, identity, data=None, record=None, **kwargs):
        """Create PID when record is created.."""
        # We create the PID after all the data has been initialized. so that
        # we can rely on having the 'id' and type set.
        self.service.record_cls.pid.create(record)


class ModelPIDComponent(PIDComponent):
    """Model PID registration component."""

    def update(self, identity, data=None, record=None, **kwargs):
        """Update an existing record."""
        record.pop("pid", None)
        record.pop("id", None)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Data Streams Celery tasks."""

from celery import shared_task

from ..datastreams import StreamEntry
from ..datastreams.factories import WriterFactory


@shared_task(ignore_result=True)
def write_entry(writer_config, entry):
    """Write an entry.

    :param writer: writer configuration as accepted by the WriterFactory.
    :param entry: dictionary, StreamEntry is not serializable.
    """
    writer = WriterFactory.create(config=writer_config)
    writer.write(StreamEntry(entry))


@shared_task(ignore_result=True)
def write_many_entry(writer_config, entries):
    """Write many entries.

    :param writer: writer configuration as accepted by the WriterFactory.
    :param entry: lisf ot dictionaries, StreamEntry is not serializable.
    """
    writer = WriterFactory.create(config=writer_config)
    stream_entries = [StreamEntry(entry) for entry in entries]
    writer.write_many(stream_entries)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2024 University of Münster.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Subjects schema tests."""

from copy import copy

import pytest
from marshmallow import ValidationError

from invenio_vocabularies.contrib.subjects.schema import SubjectSchema


def test_valid_full(app, subject_full_data, expected_subject_full_data):
    loaded = SubjectSchema().load(subject_full_data)
    assert expected_subject_full_data == loaded


def test_invalid_missing_field(app, subject_full_data):
    # no id
    invalid = copy(subject_full_data)
    del invalid["id"]
    with pytest.raises(ValidationError):
        SubjectSchema().load(invalid)

    # no scheme
    invalid = copy(subject_full_data)
    del invalid["scheme"]
    with pytest.raises(ValidationError):
        SubjectSchema().load(invalid)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024-2025 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

import io

import pytest
from rdflib import Graph

from invenio_vocabularies.contrib.subjects.nvs.datastreams import NVSSubjectsTransformer
from invenio_vocabularies.datastreams.datastreams import StreamEntry
from invenio_vocabularies.datastreams.errors import TransformerError
from invenio_vocabularies.datastreams.readers import RDFReader

VALID_XML_DATA = bytes(
    """<?xml version="1.0" encoding="UTF-8"?><rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:skos="http://www.w3.org/2004/02/skos/core#" xmlns:dc="http://purl.org/dc/terms/" xmlns:dce="http://purl.org/dc/elements/1.1/" xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#" xmlns:grg="http://www.isotc211.org/schemas/grg/" xmlns:owl="http://www.w3.org/2002/07/owl#" xmlns:void="http://rdfs.org/ns/void#" xmlns:pav="http://purl.org/pav/" xmlns:prov="https://www.w3.org/ns/prov#" xmlns:reg="http://purl.org/linked-data/registry#" xmlns:cpm="http://purl.org/voc/cpm#" xmlns:qudt="https://qudt.org/2.1/schema/qudt#" xmlns:semapv="http://w3id.org/semapv/vocab/" xmlns:iop="https://w3id.org/iadopt/ont#" xmlns:sssom="https://w3id.org/sssom/schema/" xmlns:puv="https://w3id.org/env/puv#">
  <rdf:Description rdf:about="http://vocab.nerc.ac.uk/collection/P02/current/QDMD/">
    <pav:authoredOn>2008-08-15 10:07:03.0</pav:authoredOn>
    <pav:hasCurrentVersion rdf:resource="http://vocab.nerc.ac.uk/collection/P02/current/QDMD/1/"/>
    <dce:identifier>SDN:P02::QDMD</dce:identifier>
    <pav:version>1</pav:version>
    <skos:broader rdf:resource="http://vocab.nerc.ac.uk/collection/C47/current/IN6_1_2/"/>
    <skos:broader rdf:resource="http://vocab.nerc.ac.uk/collection/P22/current/19/"/>
    <rdf:type rdf:resource="http://www.w3.org/2004/02/skos/core#Concept"/>
    <skos:definition xml:lang="en">Parameters quantifying the amount of liquids (e.g. ship's ballast) or solids (e.g. dredge spoil) deposited into the water column as a result of man's activities</skos:definition>
    <void:inDataset rdf:resource="http://vocab.nerc.ac.uk/.well-known/void"/>
    <owl:deprecated>false</owl:deprecated>
    <skos:note xml:lang="en">accepted</skos:note>
    <dc:identifier>SDN:P02::QDMD</dc:identifier>
    <skos:prefLabel xml:lang="en">Quantity of material dumped</skos:prefLabel>
    <owl:versionInfo>1</owl:versionInfo>
    <dc:date>2008-08-15 10:07:03.0</dc:date>
    <skos:altLabel>Amount_dumped</skos:altLabel>
    <skos:inScheme rdf:resource="http://vocab.nerc.ac.uk/scheme/EDMED_DCAT_THEMES/current/"/>
    <skos:notation>SDN:P02::QDMD</skos:notation>
    <skos:broader rdf:resource="http://vocab.nerc.ac.uk/collection/D01/current/D0100001/"/>
    <skos:broader rdf:resource="http://vocab.nerc.ac.uk/collection/L19/current/005/"/>
    <skos:broader rdf:resource="http://vocab.nerc.ac.uk/collection/P03/current/H001/"/>
    <skos:broader rdf:resource="http://vocab.nerc.ac.uk/collection/P05/current/007/"/>
  </rdf:Description>
</rdf:RDF>""",
    encoding="utf-8",
)

INVALID_XML_DATA = bytes(
    """<?xml version="1.0" encoding="UTF-8"?><rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:skos="http://www.w3.org/2004/02/skos/core#">
    <skos:Collection rdf:about="http://vocab.nerc.ac.uk/collection/P02/current/">
        <skos:member>
            <skos:Concept rdf:about="http://vocab.nerc.ac.uk/collection/P02/current/QDMD/">
                <skos:prefLabel>Invalid Data</skos:prefLabel>
                <!-- No skos:notation present -->
            </skos:Concept>
        </skos:member>
    </skos:Collection>
</rdf:RDF>""",
    encoding="utf-8",
)


@pytest.fixture(scope="module")
def expected_from_rdf():
    return [
        {
            "id": "SDN:P02::QDMD",
            "scheme": "NVS-P02",
            "subject": "Quantity of material dumped",
            "title": {
                "en": "Quantity of material dumped",
            },
            "props": {
                "definition": "Parameters quantifying the amount of liquids (e.g. ship's ballast) or solids (e.g. dredge spoil) deposited into the water column as a result of man's activities"
            },
            "identifiers": [
                {
                    "scheme": "url",
                    "identifier": "http://vocab.nerc.ac.uk/collection/P02/current/QDMD/",
                }
            ],
        }
    ]


def parse_rdf_data(rdf_data):
    """Parse RDF data and return a list of stream entries."""
    reader = RDFReader()
    rdf_graph = Graph()
    rdf_graph.parse(io.BytesIO(rdf_data), format="xml")
    return list(reader._iter(rdf_graph))[0]


def test_nvs_transformer_pref_label(expected_from_rdf):
    stream_entry = parse_rdf_data(VALID_XML_DATA)
    assert len(stream_entry) > 0
    transformer = NVSSubjectsTransformer()
    result = []
    entry = transformer.apply(StreamEntry(stream_entry)).entry
    result.append(entry)
    assert expected_from_rdf == result


def test_nvs_transformer_missing_id():
    stream_entry = parse_rdf_data(INVALID_XML_DATA)
    assert len(stream_entry) > 0

    transformer = NVSSubjectsTransformer()

    with pytest.raises(TransformerError) as err:
        transformer.apply(StreamEntry(stream_entry))

    expected_error = f"No id found for: {stream_entry['subject']}"

    assert expected_error in err.value.args



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
# Copyright (C) 2024 California Institute of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""ROR-related Datastreams Readers/Writers/Transformers module."""

import io

import arrow
import requests
from idutils import normalize_ror

from invenio_vocabularies.datastreams.errors import ReaderError, TransformerError
from invenio_vocabularies.datastreams.readers import BaseReader
from invenio_vocabularies.datastreams.transformers import BaseTransformer


class RORHTTPReader(BaseReader):
    """ROR HTTP Reader.

    Returning an in-memory
    binary stream of the latest ROR data dump ZIP file.
    """

    def __init__(self, origin=None, mode="r", since=None, *args, **kwargs):
        """Constructor."""
        self._since = since
        super().__init__(origin, mode, *args, **kwargs)

    def _iter(self, fp, *args, **kwargs):
        raise NotImplementedError(
            "RORHTTPReader downloads one file "
            "and therefore does not iterate through items"
        )

    def _get_last_dump_date(self, linksets):
        """Get the last dump date."""
        for linkset in linksets:
            metadata_formats = linkset.get("describedby", [])
            for format_link in metadata_formats:
                if format_link.get("type") == "application/ld+json":
                    json_ld_reponse = requests.get(
                        format_link["href"],
                        headers={"Accept": format_link["type"]},
                    )
                    json_ld_reponse.raise_for_status()
                    json_ld_data = json_ld_reponse.json()

                    last_dump_date = arrow.get(
                        json_ld_data.get("dateCreated")
                        or json_ld_data.get("datePublished")
                    )
                    return last_dump_date
        else:
            raise ReaderError(
                "Couldn't find JSON-LD in publisher's linkset "
                "to determine last dump date."
            )

    def read(self, item=None, *args, **kwargs):
        """Reads the latest ROR data dump.

        Read from ZIP file from
        Zenodo and yields an in-memory binary stream of it.
        """
        if item:
            raise NotImplementedError(
                "RORHTTPReader does not support being chained after another reader"
            )

        # Follow the DOI to get the link of the linkset
        dataset_doi_link = "https://doi.org/10.5281/zenodo.6347574"
        landing_page = requests.get(dataset_doi_link, allow_redirects=True)
        landing_page.raise_for_status()

        # Call the signposting `linkset+json` endpoint for
        # the Concept DOI (i.e. latest version) of the ROR data dump.
        # See: https://github.com/inveniosoftware/rfcs/blob/master/rfcs/rdm-0071-signposting.md#provide-an-applicationlinksetjson-endpoint
        if "linkset" not in landing_page.links:
            raise ReaderError("Linkset not found in the ROR dataset record.")
        linkset_response = requests.get(
            landing_page.links["linkset"]["url"],
            headers={"Accept": "application/linkset+json"},
        )
        linkset_response.raise_for_status()
        linksets = linkset_response.json()["linkset"]

        if self._since:
            last_dump_date = self._get_last_dump_date(linksets)
            if last_dump_date < arrow.get(self._since):
                return

        for linkset in linksets:
            items = linkset.get("item", [])
            zip_files = [item for item in items if item["type"] == "application/zip"]
            if len(zip_files) == 1:
                file_url = zip_files[0]["href"]
                break
            if len(zip_files) > 1:
                raise ReaderError(f"Expected 1 ZIP item but got {len(zip_files)}")

        # Download the ZIP file and fully load the response bytes content in memory.
        # The bytes content are then wrapped by a BytesIO to be
        # file-like object (as required by `zipfile.ZipFile`).
        # Using directly `file_resp.raw` is not possible since
        # `zipfile.ZipFile` requires the file-like object to be seekable.
        file_resp = requests.get(file_url)
        file_resp.raise_for_status()
        yield io.BytesIO(file_resp.content)


VOCABULARIES_DATASTREAM_READERS = {
    "ror-http": RORHTTPReader,
}


class RORTransformer(BaseTransformer):
    """Transforms a JSON ROR record into a funders record."""

    def __init__(
        self, *args, vocab_schemes=None, funder_fundref_doi_prefix=None, **kwargs
    ):
        """Initializes the transformer."""
        self.vocab_schemes = vocab_schemes
        self.funder_fundref_doi_prefix = funder_fundref_doi_prefix
        super().__init__(*args, **kwargs)

    def apply(self, stream_entry, **kwargs):
        """Applies the transformation to the stream entry."""
        record = stream_entry.entry
        ror = {}
        ror["title"] = {}

        ror["id"] = normalize_ror(record.get("id"))
        if not ror["id"]:
            raise TransformerError(_("Id not found in ROR entry."))

        # Using set so aliases are unique
        aliases = set()
        acronym = None
        for name in record.get("names"):
            lang = name.get("lang", "en")
            if lang == None:
                lang = "en"
            if "ror_display" in name["types"]:
                ror["name"] = name["value"]
            if "label" in name["types"]:
                ror["title"][lang] = name["value"]
            if "alias" in name["types"]:
                aliases.add(name["value"])
            if "acronym" in name["types"]:
                # The first acronyn goes in acronym field to maintain
                # compatability with existing data structure
                if not acronym:
                    acronym = name["value"]
                else:
                    aliases.add(name["value"])
        if "en" not in ror["title"]:
            ror["title"]["en"] = ror["name"]
        if acronym:
            ror["acronym"] = acronym
        if aliases:
            ror["aliases"] = list(aliases)

        # ror_display is required and should be in every entry
        if not ror["name"]:
            raise TransformerError(
                _("Name with type ror_display not found in ROR entry.")
            )

        # This only gets the first location, to maintain compatability
        # with existing data structure
        location = record.get("locations", [{}])[0].get("geonames_details", {})
        ror["country"] = location.get("country_code")
        ror["country_name"] = location.get("country_name")
        ror["location_name"] = location.get("name")

        ror["types"] = record.get("types")

        status = record.get("status")
        ror["status"] = status

        # The ROR is always listed in identifiers, expected by serialization
        ror["identifiers"] = [{"identifier": ror["id"], "scheme": "ror"}]
        if self.vocab_schemes:
            valid_schemes = set(self.vocab_schemes.keys())
        else:
            valid_schemes = set()
        fund_ref = "fundref"
        if self.funder_fundref_doi_prefix:
            valid_schemes.add(fund_ref)
        for identifier in record.get("external_ids"):
            scheme = identifier["type"]
            if scheme in valid_schemes:
                value = identifier.get("preferred") or identifier.get("all")[0]
                if scheme == fund_ref:
                    if self.funder_fundref_doi_prefix:
                        value = f"{self.funder_fundref_doi_prefix}/{value}"
                        scheme = "doi"
                ror["identifiers"].append(
                    {
                        "identifier": value,
                        "scheme": scheme,
                    }
                )

        stream_entry.entry = ror
        return stream_entry


VOCABULARIES_DATASTREAM_TRANSFORMERS = {
    "ror": RORTransformer,
}

VOCABULARIES_DATASTREAM_WRITERS = {}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test models."""

import pytest

from invenio_vocabularies.records.models import VocabularyScheme


#
# Tests
#
def test_well_formed_subtype(database):
    with pytest.raises(AssertionError):
        VocabularyScheme.create(
            id="foo,bar,baz", parent_id="subjects", name="Foo Bar Baz", uri=""
        )

    with pytest.raises(AssertionError):
        VocabularyScheme.create(
            id="foo:bar", parent_id="subjects", name="Foo: Bar", uri=""
        )

    assert VocabularyScheme.create(id="foo", parent_id="subjects", name="Foo", uri="")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024-2025 CERN.
# Copyright (C) 2024 KTH Royal Institute of Technology.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
"""Generate Vocabulary Config."""

from copy import deepcopy

import yaml
from invenio_records_resources.proxies import current_service_registry

from .contrib.affiliations.datastreams import (
    DATASTREAM_CONFIG as affiliations_ds_config,
)
from .contrib.affiliations.datastreams import (
    DATASTREAM_CONFIG_EDMO as affiliations_edmo_ds_config,
)
from .contrib.affiliations.datastreams import (
    DATASTREAM_CONFIG_OPENAIRE as affiliations_openaire_ds_config,
)
from .contrib.awards.datastreams import DATASTREAM_CONFIG as awards_ds_config
from .contrib.awards.datastreams import (
    DATASTREAM_CONFIG_CORDIS as awards_cordis_ds_config,
)
from .contrib.funders.datastreams import DATASTREAM_CONFIG as funders_ds_config
from .contrib.names.datastreams import DATASTREAM_CONFIG as names_ds_config
from .contrib.subjects.datastreams import DATASTREAM_CONFIG as subjects_ds_config
from .contrib.subjects.euroscivoc.datastreams import (
    DATASTREAM_CONFIG as euroscivoc_ds_config,
)
from .contrib.subjects.gemet.datastreams import DATASTREAM_CONFIG as gemet_ds_config
from .contrib.subjects.nvs.datastreams import DATASTREAM_CONFIG as nvs_ds_config


class VocabularyConfig:
    """Vocabulary Config Factory."""

    config = None
    vocabulary_name = None

    def get_config(self, filepath=None, origin=None):
        """Get the configuration for the vocabulary."""
        config = deepcopy(self.config)
        if filepath:
            with open(filepath, encoding="utf-8") as f:
                config = yaml.safe_load(f).get(self.vocabulary_name)
        if origin:
            config["readers"][0].setdefault("args", {})
            config["readers"][0]["args"]["origin"] = origin
        return config

    def get_service(self):
        """Get the service for the vocabulary."""
        return current_service_registry.get(self.vocabulary_name)


class NamesVocabularyConfig(VocabularyConfig):
    """Names Vocabulary Config."""

    config = names_ds_config
    vocabulary_name = "names"


class FundersVocabularyConfig(VocabularyConfig):
    """Funders Vocabulary Config."""

    config = funders_ds_config
    vocabulary_name = "funders"

    def get_service(self):
        """Get the service for the vocabulary."""
        raise NotImplementedError("Service not implemented for Funders")


class SubjectsVocabularyConfig(VocabularyConfig):
    """Subjects Vocabulary Config."""

    config = subjects_ds_config
    vocabulary_name = "subjects"

    def get_service(self):
        """Get the service for the vocabulary."""
        raise NotImplementedError("Service not implemented for Subjects")


class AwardsVocabularyConfig(VocabularyConfig):
    """Awards Vocabulary Config."""

    config = awards_ds_config
    vocabulary_name = "awards"

    def get_service(self):
        """Get the service for the vocabulary."""
        raise NotImplementedError("Service not implemented for Awards")


class AwardsCordisVocabularyConfig(VocabularyConfig):
    """Awards Vocabulary Config."""

    config = awards_cordis_ds_config
    vocabulary_name = "awards:cordis"

    def get_service(self):
        """Get the service for the vocabulary."""
        raise NotImplementedError("Service not implemented for CORDIS Awards")


class AffiliationsVocabularyConfig(VocabularyConfig):
    """Affiliations Vocabulary Config."""

    config = affiliations_ds_config
    vocabulary_name = "affiliations"

    def get_service(self):
        """Get the service for the vocabulary."""
        raise NotImplementedError("Service not implemented for Affiliations")


class AffiliationsOpenAIREVocabularyConfig(VocabularyConfig):
    """OpenAIRE Affiliations Vocabulary Config."""

    config = affiliations_openaire_ds_config
    vocabulary_name = "affiliations:openaire"

    def get_service(self):
        """Get the service for the vocabulary."""
        raise NotImplementedError("Service not implemented for OpenAIRE Affiliations")


class AffiliationsEDMOVocabularyConfig(VocabularyConfig):
    """European Directory of Marine Organisations (EDMO) Affiliations Vocabulary Config."""

    config = affiliations_edmo_ds_config
    vocabulary_name = "affiliations:edmo"

    def get_service(self):
        """Get the service for the vocabulary."""
        raise NotImplementedError("Service not implemented for EDMO Affiliations")


class SubjectsEuroSciVocVocabularyConfig(VocabularyConfig):
    """EuroSciVoc Subjects Vocabulary Config."""

    config = euroscivoc_ds_config
    vocabulary_name = "subjects:euroscivoc"

    def get_service(self):
        """Get the service for the vocabulary."""
        raise NotImplementedError("Service not implemented for EuroSciVoc Subjects")


class SubjectsGEMETVocabularyConfig(VocabularyConfig):
    """GEMET Subjects Vocabulary Config."""

    config = gemet_ds_config
    vocabulary_name = "subjects:gemet"

    def get_service(self):
        """Get the service for the vocabulary."""
        raise NotImplementedError("Service not implemented for GEMET Subjects")


class SubjectsNVSVocabularyConfig(VocabularyConfig):
    """NVS Subjects Vocabulary Config."""

    config = nvs_ds_config
    vocabulary_name = "subjects:nvs"

    def get_service(self):
        """Get the service for the vocabulary."""
        raise NotImplementedError("Service not implemented for NVS Subjects")


def get_vocabulary_config(vocabulary):
    """Factory function to get the appropriate Vocabulary Config."""
    vocab_config = {
        "names": NamesVocabularyConfig,
        "funders": FundersVocabularyConfig,
        "awards": AwardsVocabularyConfig,
        "awards:cordis": AwardsCordisVocabularyConfig,
        "affiliations": AffiliationsVocabularyConfig,
        "affiliations:openaire": AffiliationsOpenAIREVocabularyConfig,
        "affiliations:edmo": AffiliationsEDMOVocabularyConfig,
        "subjects": SubjectsVocabularyConfig,
        "subjects:gemet": SubjectsGEMETVocabularyConfig,
        "subjects:nvs": SubjectsNVSVocabularyConfig,
        "subjects:euroscivoc": SubjectsEuroSciVocVocabularyConfig,
    }
    return vocab_config.get(vocabulary, VocabularyConfig)()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the funder vocabulary resource."""

import json
from copy import deepcopy

import pytest
from invenio_db import db

from invenio_vocabularies.contrib.funders.api import Funder


@pytest.fixture(scope="module")
def prefix():
    """API prefix."""
    return "funders"


def test_funders_invalid(client, h, prefix):
    """Test invalid type."""
    # invalid type
    res = client.get(f"{prefix}/invalid", headers=h)
    assert res.status_code == 404


def test_funders_forbidden(client, h, prefix, example_funder, funder_full_data):
    """Test invalid type."""
    # invalid type
    funder_full_data_too = deepcopy(funder_full_data)
    funder_full_data_too["pid"] = "other"
    res = client.post(f"{prefix}", headers=h, data=json.dumps(funder_full_data_too))
    assert res.status_code == 403

    res = client.put(
        f"{prefix}/01ggx4157", headers=h, data=json.dumps(funder_full_data)
    )
    assert res.status_code == 403

    res = client.delete(f"{prefix}/01ggx4157")
    assert res.status_code == 403


def test_funders_get(client, example_funder, h, prefix):
    """Test the endpoint to retrieve a single item."""
    id_ = example_funder.id  # result_items wraps pid into id

    res = client.get(f"{prefix}/{id_}", headers=h)
    assert res.status_code == 200
    assert res.json["id"] == id_
    # Test links
    assert res.json["links"] == {"self": "https://127.0.0.1:5000/api/funders/01ggx4157"}


def test_funders_search(client, example_funder, h, prefix):
    """Test a successful search."""
    res = client.get(prefix, headers=h)

    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1
    assert res.json["sortBy"] == "name"


@pytest.fixture(scope="function")
def example_funders(service, identity, indexer):
    """Create dummy funders with similar names/titles."""
    funders_data = [
        {
            "id": "01ggx4157",
            "name": "CERN",
            "country": "CH",
            "title": {
                "en": "European Organization for Nuclear Research",
                "fr": "Conseil Européen pour la Recherche Nucléaire",
            },
        },
        {"id": "0aaaaaa11", "name": "OTHER", "country": "CH", "title": {"en": "CERN"}},
        {
            "id": "0aaaaaa22",
            "name": "CERT",
            "country": "CH",
            "title": {
                "en": "Computer Emergency Response Team",
                "fr": "Équipe d'Intervention d'Urgence Informatique",
            },
        },
        {
            "id": "000e0be47",
            "name": "Northwestern University",
            "country": "US",
            "title": {
                "en": "Northwestern University",
            },
        },
    ]
    funders = []
    for data in funders_data:
        funders.append(service.create(identity, data))
    Funder.index.refresh()  # Refresh the index

    yield

    for funder in funders:
        funder._record.delete(force=True)
        indexer.delete(funder._record, refresh=True)
        db.session.commit()


def test_funders_prefix_search(client, h, prefix, example_funders):
    """Test a successful search."""
    # Should show 1 result
    res = client.get(f"{prefix}?q=uni", headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1


def test_funders_suggest_sort(client, h, prefix, example_funders):
    """Test a successful search."""

    # Should show 2 results, and id=cern as first due to name
    res = client.get(f"{prefix}?suggest=CERN", headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 3
    assert res.json["hits"]["hits"][0]["name"] == "CERN"
    assert res.json["hits"]["hits"][1]["name"] == "CERT"
    # Matches lower, since title is boosted less
    assert res.json["hits"]["hits"][2]["name"] == "OTHER"
    assert res.json["hits"]["hits"][2]["title"]["en"] == "CERN"

    res = client.get(f"{prefix}?suggest=N%C3%B5rthw%C3%AAst", headers=h)  # Nõrthwêst
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1
    assert res.json["hits"]["hits"][0]["name"] == "Northwestern University"

    # Should show 0 results since scheme is not searchable
    res = client.get(f"{prefix}?suggest=nucl%C3%A9aire", headers=h)  # nucléaire
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 0


def test_funders_delete(
    client_with_credentials, h, prefix, identity, service, funder_full_data
):
    """Test a successful delete."""
    funder = service.create(identity, funder_full_data)
    id_ = funder.id
    res = client_with_credentials.delete(f"{prefix}/{id_}", headers=h)
    assert res.status_code == 204

    # only the metadata is removed from the record, it is still resolvable
    res = client_with_credentials.get(f"{prefix}/{id_}", headers=h)
    assert res.status_code == 200
    base_keys = {"created", "updated", "id", "links", "revision_id"}
    assert set(res.json.keys()) == base_keys
    # not-ideal cleanup
    funder._record.delete(force=True)


def test_funders_update(
    client_with_credentials, example_funder, funder_full_data, h, prefix
):
    """Test a successful update."""
    id_ = example_funder.id
    new_name = "updated"
    funder_full_data["name"] = new_name
    res = client_with_credentials.put(
        f"{prefix}/01ggx4157", headers=h, data=json.dumps(funder_full_data)
    )
    assert res.status_code == 200
    assert res.json["id"] == id_  # result_items wraps pid into id
    assert res.json["name"] == new_name


def test_funders_create(client_with_credentials, funder_full_data, h, prefix):
    """Tests a successful creation."""
    res = client_with_credentials.post(
        f"{prefix}", headers=h, data=json.dumps(funder_full_data)
    )
    assert res.status_code == 201
    assert res.json["id"] == funder_full_data["id"]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Names datastreams, transformers, writers and readers."""

import csv
import io
import tarfile
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import timedelta
from itertools import islice
from pathlib import Path

import arrow
import regex as re
from flask import current_app
from invenio_access.permissions import system_identity
from werkzeug.utils import cached_property

from invenio_vocabularies.contrib.names.s3client import S3OrcidClient

from ...datastreams.errors import TransformerError
from ...datastreams.readers import BaseReader, SimpleHTTPReader
from ...datastreams.transformers import BaseTransformer
from ...datastreams.writers import ServiceWriter


class OrcidDataSyncReader(BaseReader):
    """ORCiD Data Sync Reader."""

    def __init__(self, origin=None, mode="r", since=None, *args, **kwargs):
        """Constructor.

        :param origin: Data source (e.g. filepath).
                       Can be none in case of piped readers.
        """
        super().__init__(origin=origin, mode=mode, *args, **kwargs)
        self.s3_client = S3OrcidClient()
        self.since = since

    def _fetch_orcid_data(self, orcid_to_sync, bucket):
        """Fetches a single ORCiD record from S3."""
        # The ORCiD file key is located in a folder which name corresponds to the last three digits of the ORCiD
        suffix = orcid_to_sync[-3:]
        key = f"{suffix}/{orcid_to_sync}.xml"
        try:
            # Potential improvement: use the a XML jax parser to avoid loading the whole file in memory
            # and choose the sections we need to read (probably the summary)
            return self.s3_client.read_file(f"s3://{bucket}/{key}")
        except Exception:
            current_app.logger.exception("Failed to fetch ORCiD record.")

    def _process_lambda_file(self, fileobj):
        """Process the ORCiD lambda file and returns a list of ORCiDs to sync.

        The decoded fileobj looks like the following:
        orcid, path, date_created, last_modified
        0000-0001-5109-3700, http://orcid.org/0000-0001-5109-3700, 2014-08-02 15:00:00.000,2021-08-02 15:00:00.000

        Yield ORCiDs to sync until the last sync date is reached.
        """
        date_format = "YYYY-MM-DD HH:mm:ss.SSSSSS"
        date_format_no_millis = "YYYY-MM-DD HH:mm:ss"
        time_shift = current_app.config["VOCABULARIES_ORCID_SYNC_SINCE"]
        if self.since:
            time_shift = self.since
        last_sync = arrow.now() - timedelta(**time_shift)
        try:
            content = io.TextIOWrapper(fileobj, encoding="utf-8")
            csv_reader = csv.DictReader(content)

            for row in csv_reader:  # Skip the header line
                orcid = row["orcid"]

                # Lambda file is ordered by last modified date
                last_modified_str = row["last_modified"]
                try:
                    last_modified_date = arrow.get(last_modified_str, date_format)
                except arrow.parser.ParserError:
                    last_modified_date = arrow.get(
                        last_modified_str, date_format_no_millis
                    )

                if last_modified_date < last_sync:
                    break
                yield orcid
        finally:
            fileobj.close()

    def _iter(self, orcids):
        """Iterates over the ORCiD records yielding each one."""
        with ThreadPoolExecutor(
            max_workers=current_app.config["VOCABULARIES_ORCID_SYNC_MAX_WORKERS"]
        ) as executor:
            # futures is a dictionary where the key is the ORCID value and the item is the Future object
            futures = {
                orcid: executor.submit(
                    self._fetch_orcid_data,
                    orcid,
                    current_app.config["VOCABULARIES_ORCID_SUMMARIES_BUCKET"],
                )
                for orcid in orcids
            }

            for orcid in list(futures.keys()):
                try:
                    result = futures[orcid].result()
                    if result:
                        yield result
                finally:
                    # Explicitly release memory, as we don't need the future anymore.
                    # This is mostly required because as long as we keep a reference to the future
                    # (in the above futures dict), the garbage collector won't collect it
                    # and it will keep the memory allocated.
                    del futures[orcid]

    def read(self, item=None, *args, **kwargs):
        """Streams the ORCiD lambda file, process it to get the ORCiDS to sync and yields it's data."""
        # Read the file from S3
        tar_content = self.s3_client.read_file(
            "s3://orcid-lambda-file/last_modified.csv.tar"
        )

        # Opens tar file and process it
        with tarfile.open(fileobj=io.BytesIO(tar_content)) as tar:
            # Iterate over each member (file or directory) in the tar file
            for member in tar.getmembers():
                # Extract the file
                extracted_file = tar.extractfile(member)
                if extracted_file:
                    current_app.logger.info(f"[ORCID Reader] Processing lambda file...")
                    # Process the file and get the ORCiDs to sync
                    orcids_to_sync = set(self._process_lambda_file(extracted_file))

                    # Close the file explicitly after processing
                    extracted_file.close()

                    # Process ORCIDs in smaller batches
                    for orcid_batch in self._chunked_iter(
                        orcids_to_sync, batch_size=100
                    ):
                        yield from self._iter(orcid_batch)

    def _chunked_iter(self, iterable, batch_size):
        """Yield successive chunks of a given size."""
        it = iter(iterable)
        while chunk := list(islice(it, batch_size)):
            yield chunk


class OrcidHTTPReader(SimpleHTTPReader):
    """ORCiD HTTP Reader."""

    def __init__(self, *args, test_mode=True, **kwargs):
        """Constructor."""
        if test_mode:
            origin = "https://sandbox.orcid.org/{id}"
        else:
            origin = "https://orcid.org/{id}"

        super().__init__(origin, *args, **kwargs)


DEFAULT_NAMES_EXCLUDE_REGEX = r"[\p{P}\p{S}\p{Nd}\p{No}\p{Emoji}--,.()\-']"
"""Regex to filter out names with punctuation, symbols, numbers and emojis."""


class OrcidOrgToAffiliationMapper:
    """Default ORCiD Org ID to affiliation ID mapper."""

    def __init__(self, org_ids_mapping=None, org_ids_mapping_file=None):
        """Constructor."""
        self._org_ids_mapping = org_ids_mapping
        self._org_ids_mapping_file = org_ids_mapping_file

    @cached_property
    def org_ids_mapping(self):
        """Mapping of ORCiD org IDs to affiliation IDs."""
        org_ids_mapping_file = self._org_ids_mapping_file or current_app.config.get(
            "VOCABULARIES_ORCID_ORG_IDS_MAPPING_PATH"
        )
        if org_ids_mapping_file:
            org_ids_mapping_file = Path(org_ids_mapping_file)
            # If the path is relative, prepend the instance path
            if not org_ids_mapping_file.is_absolute():
                org_ids_mapping_file = (
                    Path(current_app.instance_path) / org_ids_mapping_file
                )
            with open(org_ids_mapping_file) as fin:
                result = {}
                reader = csv.reader(fin)

                # Check if the first row is a header
                org_scheme, org_id, aff_id = next(reader)
                if org_scheme.lower() != "org_scheme":
                    result[(org_scheme, org_id)] = aff_id

                for org_scheme, org_id, aff_id in reader:
                    result[(org_scheme, org_id)] = aff_id

                return result

        return self._org_ids_mapping or {}

    def __call__(self, org_scheme, org_id):
        """Map an ORCiD org ID to an affiliation ID."""
        # By default we know that ROR IDs are linkable
        if org_scheme == "ROR":
            return org_id.split("/")[-1]
        # Otherwise see if we have a mapping from other schemes to an affiliation ID
        return self.org_ids_mapping.get((org_scheme, org_id))


class OrcidTransformer(BaseTransformer):
    """Transforms an ORCiD record into a names record."""

    def __init__(
        self,
        *args,
        names_exclude_regex=DEFAULT_NAMES_EXCLUDE_REGEX,
        org_id_to_affiliation_id_func=None,
        **kwargs,
    ) -> None:
        """Constructor."""
        self._names_exclude_regex = names_exclude_regex
        self._org_id_to_affiliation_id_func = (
            org_id_to_affiliation_id_func or OrcidOrgToAffiliationMapper()
        )
        super().__init__()

    def org_id_to_affiliation_id(self, org_scheme, org_id):
        """Convert and ORCiD org ID to a linkable affiliation ID."""
        return self._org_id_to_affiliation_id_func(org_scheme, org_id)

    def apply(self, stream_entry, **kwargs):
        """Applies the transformation to the stream entry."""
        record = stream_entry.entry
        person = record["person"]
        orcid_id = record["orcid-identifier"]["path"]

        name = person.get("name")
        if name is None:
            raise TransformerError("Name not found in ORCiD entry.")
        if name.get("family-name") is None:
            raise TransformerError("Family name not found in ORCiD entry.")

        if not self._is_valid_name(name["given-names"] + name["family-name"]):
            raise TransformerError("Invalid characters in name.")

        entry = {
            "id": orcid_id,
            "given_name": name.get("given-names"),
            "family_name": name.get("family-name"),
            "identifiers": [{"scheme": "orcid", "identifier": orcid_id}],
            "affiliations": self._extract_affiliations(record),
        }

        stream_entry.entry = entry
        return stream_entry

    def _is_valid_name(self, name):
        """Check whether the name passes the regex."""
        if not self._names_exclude_regex:
            return True
        return not bool(re.search(self._names_exclude_regex, name, re.UNICODE | re.V1))

    def _extract_affiliations(self, record):
        """Extract affiliations from the ORCiD record."""
        result = []
        try:
            employments = (
                record.get("activities-summary", {})
                .get("employments", {})
                .get("affiliation-group", [])
            )

            # If there are single values, the XML to dict, doesn't wrap them in a list
            if isinstance(employments, dict):
                employments = [employments]

            # Remove the "employment-summary" nesting
            employments = [
                employment.get("employment-summary", {}) for employment in employments
            ]

            for employment in employments:
                terminated = employment.get("end-date")
                if terminated:
                    continue

                org = employment["organization"]
                aff_id = self._extract_affiliation_id(org)

                # Skip adding if the ID already exists in result
                if aff_id and any(aff.get("id") == aff_id for aff in result):
                    continue

                # Skip adding if the name exists in result with no ID
                if any(
                    aff.get("name") == org["name"] and "id" not in aff for aff in result
                ):
                    continue

                aff = {"name": org["name"]}
                if aff_id:
                    aff["id"] = aff_id

                result.append(aff)
        except Exception:
            pass
        return result

    def _extract_affiliation_id(self, org):
        """Extract the affiliation ID from an ORCiD organization."""
        dis_org = org.get("disambiguated-organization")
        if not dis_org:
            return

        aff_id = None
        org_id = dis_org.get("disambiguated-organization-identifier")
        org_scheme = dis_org.get("disambiguation-source")
        if org_id and org_scheme:
            aff_id = self.org_id_to_affiliation_id(org_scheme, org_id)
        return aff_id


class NamesServiceWriter(ServiceWriter):
    """Names service writer."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        service_or_name = kwargs.pop("service_or_name", "names")
        super().__init__(service_or_name=service_or_name, *args, **kwargs)

    def _entry_id(self, entry):
        """Get the id from an entry."""
        return entry["id"]


VOCABULARIES_DATASTREAM_READERS = {
    "orcid-http": OrcidHTTPReader,
    "orcid-data-sync": OrcidDataSyncReader,
}


VOCABULARIES_DATASTREAM_TRANSFORMERS = {
    "orcid": OrcidTransformer,
}
"""ORCiD Data Streams transformers."""


VOCABULARIES_DATASTREAM_WRITERS = {
    "names-service": NamesServiceWriter,
}
"""ORCiD Data Streams transformers."""


DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "tar",
            "args": {
                "regex": "\\.xml$",
            },
        },
        {
            "type": "xml",
            "args": {
                "root_element": "record",
            },
        },
    ],
    "transformers": [{"type": "orcid"}],
    "writers": [
        {
            "type": "names-service",
            "args": {
                "identity": system_identity,
            },
        }
    ],
}
"""ORCiD Data Stream configuration.

An origin is required for the reader.
"""

# TODO: Used on the jobs and should be set as a "PRESET" (naming to be defined)
ORCID_PRESET_DATASTREAM_CONFIG = {
    "readers": [
        {
            "type": "orcid-data-sync",
        },
        {
            "type": "xml",
            "args": {
                "root_element": "record",
            },
        },
    ],
    "transformers": [{"type": "orcid"}],
    "writers": [
        {
            "type": "async",
            "args": {
                "writer": {
                    "type": "names-service",
                }
            },
        }
    ],
    "batch_size": 1000,
    "write_many": True,
}
"""ORCiD Data Stream configuration.

An origin is required for the reader.
"""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test the award vocabulary resource."""

import json
from copy import deepcopy

import pytest
from invenio_db import db

from invenio_vocabularies.contrib.awards.api import Award


@pytest.fixture(scope="module")
def prefix():
    """API prefix."""
    return "awards"


def test_awards_invalid(client, h, prefix):
    """Test invalid type."""
    # invalid type
    res = client.get(f"{prefix}/invalid", headers=h)
    assert res.status_code == 404


def test_awards_forbidden(client, h, prefix, example_award, award_full_data):
    """Test invalid type."""
    # invalid type
    award_full_data_too = deepcopy(award_full_data)
    award_full_data_too["pid"] = "other"
    res = client.post(f"{prefix}", headers=h, data=json.dumps(award_full_data_too))
    assert res.status_code == 403

    res = client.put(f"{prefix}/755021", headers=h, data=json.dumps(award_full_data))
    assert res.status_code == 403

    res = client.delete(f"{prefix}/755021")
    assert res.status_code == 403


def test_awards_get(client, example_award, h, prefix):
    """Test the endpoint to retrieve a single item."""
    id_ = example_award.id  # result_items wraps pid into id

    res = client.get(f"{prefix}/{id_}", headers=h)
    assert res.status_code == 200
    assert res.json["id"] == id_
    # Test links
    assert res.json["links"] == {"self": "https://127.0.0.1:5000/api/awards/755021"}


def test_awards_search(client, example_award, h, prefix):
    """Test a successful search."""
    res = client.get(prefix, headers=h)

    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1
    assert res.json["sortBy"] == "newest"
    assert res.json["aggregations"]["funders"]

    funders_agg = res.json["aggregations"]["funders"]["buckets"][0]
    assert funders_agg["key"] == "00k4n6c32"
    assert funders_agg["doc_count"] == 1
    assert funders_agg["label"] == "European Commission (BE)"

    res = client.get(f"{prefix}?q=755021", headers=h)

    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1
    assert res.json["sortBy"] == "bestmatch"


@pytest.fixture(scope="function")
def example_awards(service, identity, indexer, example_funder_ec):
    """Create dummy awards with similar ids/numbers/titles."""
    awards_data = [
        {
            "title": {
                "en": "Host directed medicine in invasive fungal infection",
            },
            "id": "847507",
            "number": "847507",
        },
        {
            "title": {
                "en": "Palliative care in Parkinson disease",
            },
            "id": "825785",
            "number": "825785",
            "funder": {"id": example_funder_ec.id},
        },
        {
            "title": {
                "en": "Palliative Show",
            },
            "acronym": "Palliative",
            "id": "000001",
            "number": "000001",
            "funder": {"name": "Another Funder"},
        },
    ]

    awards = []
    for data in awards_data:
        awards.append(service.create(identity, data))

    Award.index.refresh()  # Refresh the index

    yield

    for award in awards:
        award._record.delete(force=True)
        indexer.delete(award._record, refresh=True)
        db.session.commit()


def test_awards_suggest_sort(client, h, prefix, example_awards):
    """Test a successful search."""
    # Should show 1 result
    res = client.get(f"{prefix}?suggest=847507", headers=h)
    assert res.status_code == 200
    assert res.json["hits"]["total"] == 1
    assert res.json["hits"]["hits"][0]["id"] == "847507"

    # Should show 1 result
    res = client.get(f"{prefix}?suggest=Parkin", headers=h)
    assert res.status_code == 200
    assert (
        res.json["hits"]["total"] == 0
    )  # should be 1 , TODO - suggestions were failing due to too many languages
    # assert res.json["hits"]["hits"][0]["id"] == "825785"

    # Should show 2 results, but pid=847507 as first due to created date
    res = client.get(f"{prefix}?suggest=Palliative", headers=h)
    assert res.status_code == 200
    assert (
        res.json["hits"]["total"] == 1
    )  # should be 2 , TODO - suggestions were failing due to too many languages
    assert res.json["hits"]["hits"][0]["id"] == "000001"
    # assert res.json["hits"]["hits"][1]["id"] == "825785"


def test_awards_faceted_suggest(client, h, prefix, example_funder_ec, example_awards):
    """Test a successful suggest with filtering."""
    # Should show 1 results because of the funder filtering
    res = client.get(
        f"{prefix}?funders={example_funder_ec.id}&suggest=Palliative",
        headers=h,
    )
    assert res.status_code == 200
    assert (
        res.json["hits"]["total"] == 0
    )  # should be 1 , TODO - suggestions were failing due to too many languages
    # assert res.json["hits"]["hits"][0]["id"] == "825785"


def test_awards_delete(
    client_with_credentials,
    h,
    prefix,
    identity,
    service,
    award_full_data,
    example_funder_ec,
):
    """Test a successful delete."""
    award = service.create(identity, award_full_data)
    id_ = award.id
    res = client_with_credentials.delete(f"{prefix}/{id_}", headers=h)
    assert res.status_code == 204

    # only the metadata is removed from the record, it is still resolvable
    res = client_with_credentials.get(f"{prefix}/{id_}", headers=h)
    assert res.status_code == 200
    base_keys = {"created", "updated", "id", "links", "revision_id"}
    assert set(res.json.keys()) == base_keys
    # not-ideal cleanup
    award._record.delete(force=True)


def test_awards_update(
    client_with_credentials, example_award, award_full_data, h, prefix
):
    """Test a successful update."""
    id_ = example_award.id
    new_title = "updated"
    award_full_data["title"]["en"] = new_title
    res = client_with_credentials.put(
        f"{prefix}/755021", headers=h, data=json.dumps(award_full_data)
    )
    assert res.status_code == 200
    assert res.json["id"] == id_  # result_items wraps pid into id
    assert res.json["title"]["en"] == new_title


def test_awards_create(
    client_with_credentials, award_full_data, h, prefix, example_funder_ec
):
    """Tests a successful creation."""
    res = client_with_credentials.post(
        f"{prefix}", headers=h, data=json.dumps(award_full_data)
    )
    assert res.status_code == 201
    assert res.json["id"] == award_full_data["id"]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary awards."""

from .awards import record_type

AwardsServiceConfig = record_type.service_config_cls

AwardsService = record_type.service_cls



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabulary awards."""

from flask_resources import (
    BaseListSchema,
    JSONSerializer,
    MarshmallowSerializer,
    ResponseHandler,
)
from invenio_db import db
from invenio_records.dumpers import SearchDumper
from invenio_records.dumpers.indexedat import IndexedAtDumperExt
from invenio_records.dumpers.relations import RelationDumperExt
from invenio_records.systemfields import MultiRelationsField
from invenio_records_resources.factories.factory import RecordTypeFactory
from invenio_records_resources.records.systemfields import (
    ModelPIDField,
    PIDListRelation,
    PIDRelation,
)
from invenio_records_resources.resources.records.headers import etag_headers

from invenio_vocabularies.services.permissions import PermissionPolicy

from ..funders.api import Funder
from ..subjects.api import Subject
from .config import AwardsSearchOptions, service_components
from .schema import AwardSchema
from .serializer import AwardL10NItemSchema

award_relations = MultiRelationsField(
    funders=PIDRelation(
        "funder",
        keys=["name"],
        pid_field=Funder.pid,
        cache_key="funder",
    ),
    subjects=PIDListRelation(
        "subjects",
        keys=["subject", "scheme", "identifiers", "props"],
        pid_field=Subject.pid,
        cache_key="subjects",
    ),
)

record_type = RecordTypeFactory(
    "Award",
    # Data layer
    pid_field_cls=ModelPIDField,
    pid_field_kwargs={
        "model_field_name": "pid",
    },
    model_cls_attrs={
        # cannot set to nullable=False because it would fail at
        # service level when create({}), see records-resources.
        "pid": db.Column(db.String(255), unique=True),
    },
    record_dumper=SearchDumper(
        model_fields={"pid": ("id", str)},
        extensions=[
            RelationDumperExt("relations"),
            IndexedAtDumperExt(),
        ],
    ),
    schema_version="1.0.0",
    schema_path="local://awards/award-v1.0.0.json",
    record_relations=award_relations,
    # Service layer
    service_id="awards",
    service_schema=AwardSchema,
    search_options=AwardsSearchOptions,
    service_components=service_components,
    permission_policy_cls=PermissionPolicy,
    # Resource layer
    endpoint_route="/awards",
    resource_cls_attrs={
        "response_handlers": {
            "application/json": ResponseHandler(JSONSerializer(), headers=etag_headers),
            "application/vnd.inveniordm.v1+json": ResponseHandler(
                MarshmallowSerializer(
                    format_serializer_cls=JSONSerializer,
                    object_schema_cls=AwardL10NItemSchema,
                    list_schema_cls=BaseListSchema,
                ),
                headers=etag_headers,
            ),
        }
    },
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 CERN.
#
# Invenio-RDM-Records is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Custom Fields for InvenioRDM."""

from invenio_records_resources.records.systemfields import PIDListRelation, PIDRelation
from invenio_records_resources.services.custom_fields.base import BaseCF
from marshmallow import fields

from ...proxies import current_service
from ...records.api import Vocabulary
from ...resources.serializer import VocabularyL10NItemSchema
from ...services.schema import VocabularyRelationSchema


class VocabularyCF(BaseCF):
    """Vocabulary custom field.

    Supporting common vocabulary structure.
    """

    field_keys = ["id", "props", "title", "icon"]
    """Return field's keys for querying.

    These keys are used to select which information to return from the
    vocabulary that is queried.
    """

    def __init__(
        self,
        name,
        vocabulary_id,
        multiple=False,
        dump_options=True,
        sort_by=None,
        schema=VocabularyRelationSchema,
        ui_schema=VocabularyL10NItemSchema,
        **kwargs,
    ):
        """Constructor."""
        super().__init__(name, **kwargs)
        self.relation_cls = PIDRelation if not multiple else PIDListRelation
        self.vocabulary_id = vocabulary_id
        self.dump_options = dump_options
        self.multiple = multiple
        self.sort_by = sort_by
        self.schema = schema
        self.ui_schema = ui_schema
        self.pid_field = Vocabulary.pid.with_type_ctx(self.vocabulary_id)

    @property
    def mapping(self):
        """Return the mapping."""
        _mapping = {
            "type": "object",
            "properties": {
                "@v": {"type": "keyword"},
                "id": {"type": "keyword"},
                "title": {"type": "object", "dynamic": "true"},
            },
        }

        return _mapping

    @property
    def field(self):
        """Marshmallow schema for vocabulary custom fields."""
        return fields.Nested(self.schema, many=self.multiple, **self._field_args)

    @property
    def ui_field(self):
        """Marshmallow UI schema for vocabulary custom fields.

        This schema is used in the UIJSONSerializer and controls how the field will be
        dumped in the UI. It takes responsibility of the localization of strings.
        """
        return fields.Nested(self.ui_schema, many=self.multiple, **self._field_args)

    def options(self, identity):
        """Return UI serialized vocabulary items."""
        if self.dump_options:
            vocabs = current_service.read_all(
                identity,
                fields=self.field_keys,
                type=self.vocabulary_id,
                sort=self.sort_by,
            )
            options = []
            for vocab in vocabs:
                options.append(VocabularyL10NItemSchema().dump(vocab))

            return options



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
# Copyright (C) 2024 Uni Münster.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Vocabularies admin interface."""

from invenio_administration.views.base import (
    AdminResourceEditView,
    AdminResourceListView,
)


class VocabulariesListView(AdminResourceListView):
    """Configuration for vocabularies list view."""

    api_endpoint = "/vocabularies/"
    name = "vocabulary-types"
    menu_label = "Vocabulary Types"
    resource_config = "vocabulary_admin_resource"
    search_request_headers = {"Accept": "application/json"}
    title = "Vocabulary Types"
    category = "Site management"

    pid_path = "id"
    icon = "exchange"
    template = "invenio_administration/search.html"

    display_search = True
    display_delete = False
    display_edit = False
    display_create = False

    item_field_list = {
        "id": {"text": "Name", "order": 1},
        "count": {"text": "Number of entries", "order": 2},
    }

    search_config_name = "VOCABULARIES_TYPES_SEARCH"
    search_facets_config_name = "VOCABULARIES_TYPES_FACETS"
    search_sort_config_name = "VOCABULARIES_TYPES_SORT_OPTIONS"



#
# This file is part of Invenio.
# Copyright (C) 2024 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Names service components."""

from invenio_records_resources.services.records.components import ServiceComponent


class InternalIDComponent(ServiceComponent):
    """Service component for internal id field."""

    field = "internal_id"

    def create(self, identity, data=None, record=None, **kwargs):
        """Create handler."""
        setattr(record, self.field, data.pop(self.field, None))

    def update(self, identity, data=None, record=None, **kwargs):
        """Update handler."""
        setattr(record, self.field, data.pop(self.field, None))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Datastream errors."""


class ReaderError(Exception):
    """Transformer application exception."""


class TransformerError(Exception):
    """Transformer application exception."""


class WriterError(Exception):
    """Transformer application exception."""


class FactoryError(Exception):
    """Transformer application exception."""

    def __init__(self, name, key):
        """Initialise error."""
        super().__init__(f"{name} {key} not configured.")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Awards JSONSchema tests."""

import pytest
from jsonschema.exceptions import ValidationError

from invenio_vocabularies.contrib.awards.api import Award


@pytest.fixture(scope="module")
def schema():
    """Returns the schema location."""
    return "local://awards/award-v1.0.0.json"


def validates(data):
    """Validates award data."""
    Award(data).validate()

    return True


def fails(data):
    """Validates award data."""
    pytest.raises(ValidationError, validates, data)
    return True


def test_valid_full(appctx, schema):
    data = {
        "$schema": schema,
        "identifiers": [
            {
                "identifier": "https://cordis.europa.eu/project/id/755021",
                "scheme": "url",
            }
        ],
        "title": {
            "en": "Personalised Treatment For Cystic Fibrosis Patients With \
                Ultra-rare CFTR Mutations (and beyond)"
        },
        "number": "755021",
        "funder": {"id": "ria", "name": "Research annd Innovation action"},
    }

    assert validates(data)


def test_valid_empty(appctx, schema):
    # check there are no requirements at JSONSchema level
    data = {"$schema": schema}

    assert validates(data)


# only number is defined by the award schema
# the rest are inherited and should be tested elsewhere


def test_fails_number(appctx, schema):
    data = {"$schema": schema, "number": 123}

    assert fails(data)
